/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Hash;
import fr.x9c.cadmium.support.Helper;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents an attribute of a Java element.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.1
 * @since 1.0
 */
public abstract class Attribute implements ToValue {

    /** Hash value for 'AnnotationDefault' attribute. */
    public static final Value HASH_ANNOTATION_DEFAULT =
        Hash.hashVariant("AnnotationDefault");

    /** Hash value for 'Code' attribute. */
    public static final Value HASH_CODE =
        Hash.hashVariant("Code");

    /** Hash value for 'ConstantValue' attribute. */
    public static final Value HASH_CONSTANT_VALUE =
        Hash.hashVariant("ConstantValue");

    /** Hash value for 'Deprecated' attribute. */
    public static final Value HASH_DEPRECATED =
        Hash.hashVariant("Deprecated");

    /** Hash value for 'EnclosingMethod' attribute. */
    public static final Value HASH_ENCLOSING_METHOD =
        Hash.hashVariant("EnclosingMethod");

    /** Hash value for 'Exceptions' attribute. */
    public static final Value HASH_EXCEPTIONS =
        Hash.hashVariant("Exceptions");

    /** Hash value for 'InnerClasses' attribute. */
    public static final Value HASH_INNER_CLASSES =
        Hash.hashVariant("InnerClasses");

    /** Hash value for 'LineNumberTable' attribute. */
    public static final Value HASH_LINE_NUMBER_TABLE =
        Hash.hashVariant("LineNumberTable");

    /** Hash value for 'LocalVariableTable' attribute. */
    public static final Value HASH_LOCAL_VARIABLE_TABLE =
        Hash.hashVariant("LocalVariableTable");

    /** Hash value for 'LocalVariableTypeTable' attribute. */
    public static final Value HASH_LOCAL_VARIABLE_TYPE_TABLE =
        Hash.hashVariant("LocalVariableTypeTable");

    /** Hash value for 'RuntimeInvisibleAnnotations' attribute. */
    public static final Value HASH_RUNTIME_INVISIBLE_ANNOTATIONS =
        Hash.hashVariant("RuntimeInvisibleAnnotations");

    /** Hash value for 'RuntimeInvisibleParameterAnnotations' attribute. */
    public static final Value HASH_RUNTIME_INVISIBLE_PARAMETER_ANNOTATIONS =
        Hash.hashVariant("RuntimeInvisibleParameterAnnotations");

    /** Hash value for 'RuntimeVisibleAnnotations' attribute. */
    public static final Value HASH_RUNTIME_VISIBLE_ANNOTATIONS =
        Hash.hashVariant("RuntimeVisibleAnnotations");

    /** Hash value for 'RuntimeVisibleParameterAnnotations' attribute. */
    public static final Value HASH_RUNTIME_VISIBLE_PARAMETER_ANNOTATIONS =
        Hash.hashVariant("RuntimeVisibleParameterAnnotations");

    /** Hash value for 'Signature' attribute. */
    public static final Value HASH_SIGNATURE =
        Hash.hashVariant("Signature");

    /** Hash value for 'SourceDebugExtension' attribute. */
    public static final Value HASH_SOURCE_DEBUG_EXTENSION =
        Hash.hashVariant("SourceDebugExtension");

    /** Hash value for 'SourceFile' attribute. */
    public static final Value HASH_SOURCE_FILE =
        Hash.hashVariant("SourceFile");

    /** Hash value for 'Synthetic' attribute. */
    public static final Value HASH_SYNTHETIC =
        Hash.hashVariant("Synthetic");

    /** Hash value for 'Synthetic' attribute. */
    public static final Value HASH_STACK_MAP_TABLE =
        Hash.hashVariant("StackMapTable");

    /** Hash value for 'Unknown' attribute. */
    public static final Value HASH_UNKNOWN =
        Hash.hashVariant("Unknown");

    /**
     * Parent class of all constant values embedded in attributes.
     */
    public static abstract class ConstantVal implements ToValue {

        /**
         * Ensures that no constant value could be defined outside this file.
         */
        private ConstantVal() {
        } // end empty constructor

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        static ConstantVal fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            switch (b.getTag()) {
            case 0: return new LongValue(b.get(0).asBlock().asInt64());
            case 1: return new FloatValue((float) b.get(0).asBlock().asDouble());
            case 2: return new DoubleValue(b.get(0).asBlock().asDouble());
            case 3: return new BooleanValue(b.get(0) == Value.TRUE);
            case 4: return new ByteValue((byte) b.get(0).asLong());
            case 5: return new CharValue((char) b.get(0).asLong());
            case 6: return new ShortValue((short) b.get(0).asLong());
            case 7: return new IntegerValue(b.get(0).asBlock().asInt32());
            case 8: return new StringValue(API.stringOfUTF8(b.get(0)));
            default:
                assert false : "invalid tag";
                return null;
            } // end switch
        } // end method 'fromValue(Value)'

    } // end inner-class 'ConstantVal'

    /**
     * Represents a long value embedded inside an attribute.
     */
    public static final class LongValue extends ConstantVal {

        /** The underlying value. */
        private final long value;

        /**
         * Constructs a long value.
         * @param v value
         */
        public LongValue(final long v) {
            this.value = v;
        } // end constructor(long)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public long getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0, Helper.createInt64(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "LongValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof LongValue) {
                final LongValue that = (LongValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LongValue'

    /**
     * Represents a float value embedded inside an attribute.
     */
    public static final class FloatValue extends ConstantVal {

        /** The underlying value. */
        private final float value;

        /**
         * Constructs a float value.
         * @param v value
         */
        public FloatValue(final float v) {
            this.value = v;
        } // end constructor(float)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public float getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(1, Helper.createFloat(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "FloatValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof FloatValue) {
                final FloatValue that = (FloatValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'FloatValue'

    /**
     * Represents a double value embedded inside an attribute.
     */
    public static final class DoubleValue extends ConstantVal {

        /** The underlying value. */
        private final double value;

        /**
         * Constructs a double value.
         * @param v value
         */
        public DoubleValue(final double v) {
            this.value = v;
        } // end constructor(double)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public double getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(2, Helper.createFloat(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "DoubleValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return (int) this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof DoubleValue) {
                final DoubleValue that = (DoubleValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'DoubleValue'

    /**
     * Represents a boolean value embedded inside an attribute.
     */
    public static final class BooleanValue extends ConstantVal {

        /** The underlying value. */
        private final boolean value;

        /**
         * Constructs a boolean value.
         * @param v value
         */
        public BooleanValue(final boolean v) {
            this.value = v;
        } // end constructor(boolean)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public boolean getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(3, this.value ? Value.TRUE : Value.FALSE);
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "BooleanValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value ? 1 : 0;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof BooleanValue) {
                final BooleanValue that = (BooleanValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'BooleanValue'

    /**
     * Represents a byte value embedded inside an attribute.
     */
    public static final class ByteValue extends ConstantVal {

        /** The underlying value. */
        private final byte value;

        /**
         * Constructs a byte value.
         * @param v value
         */
        public ByteValue(final byte v) {
            this.value = v;
        } // end constructor(byte)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public byte getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(4, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "ByteValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof ByteValue) {
                final ByteValue that = (ByteValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ByteValue'

    /**
     * Represents a char value embedded inside an attribute.
     */
    public static final class CharValue extends ConstantVal {

        /** The underlying value. */
        private final char value;

        /**
         * Constructs a char value.
         * @param v value
         */
        public CharValue(final char v) {
            this.value = v;
        } // end constructor(char)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public char getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(5, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "CharValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof CharValue) {
                final CharValue that = (CharValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'CharValue'

    /**
     * Represents a short value embedded inside an attribute.
     */
    public static final class ShortValue extends ConstantVal {

        /** The underlying value. */
        private final short value;

        /**
         * Constructs a short value.
         * @param v value
         */
        public ShortValue(final short v) {
            this.value = v;
        } // end constructor(short)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public short getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(6, Value.createFromLong(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "ShortValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof ShortValue) {
                final ShortValue that = (ShortValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ShortValue'

    /**
     * Represents a integer value embedded inside an attribute.
     */
    public static final class IntegerValue extends ConstantVal {

        /** The underlying value. */
        private final int value;

        /**
         * Constructs a integer value.
         * @param v value
         */
        public IntegerValue(final int v) {
            this.value = v;
        } // end constructor(integer)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public int getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(7, Helper.createInt32(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "IntegerValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value;
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof IntegerValue) {
                final IntegerValue that = (IntegerValue) obj;
                return this.value == that.value;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'IntegerValue'

    /**
     * Represents a string value embedded inside an attribute.
     */
    public static final class StringValue extends ConstantVal {

        /** The underlying value. */
        private final String value;

        /**
         * Constructs a string value.
         * @param v value - should not be <tt>null</tt>
         */
        public StringValue(final String v) {
            assert v != null : "null v";
            this.value = v;
        } // end constructor(String)

        /**
         * Returns the underlying value.
         * @return the underlying value
         */
        public String getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(8, API.createUTF8(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "StringValue(" + this.value + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.value.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof StringValue) {
                final StringValue that = (StringValue) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'StringValue'

    /**
     * This class implements elements to be used in an exception table.
     */
    public static final class ExceptionTableElement implements ToValue {

        /** The starting offset of protected code. */
        private final int start;

        /** The ending offset of protected code. */
        private final int end;

        /** The handler offset. */
        private final int handler;

        /** The exception to be caught (<tt>null</tt> means all). */
        private final String exception;

        /**
         * Constructs an exception table element from offsets and name.
         * @param s starting offset of protected code - should be in 0..65535
         * @param e ending offset of protected code - should be in 0..65535
         * @param h handler offset - should be in 0..65535
         * @param exn exception to be caught (<tt>null</tt> means all)
         */
        public ExceptionTableElement(final int s,
                                     final int e,
                                     final int h,
                                     final String exn) {
            assert (s >= 0) && (s <= 65535) : "invalid s";
            assert (e >= 0) && (e <= 65535) : "invalid e";
            assert (h >= 0) && (h <= 65535) : "invalid h";
            this.start = s;
            this.end = e;
            this.handler = h;
            this.exception = exn;
        } // end constructor(int, int, int, String)

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Value exn;
            if (this.exception == null) {
                exn = Value.ZERO;
            } else {
                exn = Value.createFromBlock(Block.createBlock(0, API.createClassName(this.exception)));
            } // end if/else
            final Block b = Block.createBlock(0,
                                              Value.createFromLong(this.start),
                                              Value.createFromLong(this.end),
                                              Value.createFromLong(this.handler),
                                              exn);
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        static ExceptionTableElement fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            return new ExceptionTableElement(b.get(0).asLong(),
                                             b.get(1).asLong(),
                                             b.get(2).asLong(),
                                             b.get(3) == Value.ZERO
                                             ? null
                                             : API.stringOfClassName(b.get(3).asBlock().get(0)));
        } // end method 'fromValue(Value)'

        /**
         * {@inheritDoc}
         */
        @Override
        public final String toString() {
            return "ExceptionTableElement("
                + this.start + ", "
                + this.end + ", "
                + this.handler + ", "
                + this.exception + ")";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final int hashCode() {
            return this.start
                + this.end
                + this.handler
                + (this.exception != null ? this.exception.hashCode() : 0);
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof ExceptionTableElement) {
                final ExceptionTableElement that = (ExceptionTableElement) obj;
                return (this.start == that.start)
                    && (this.end == that.end)
                    && (this.handler == that.handler)
                    && API.nullOrEquals(this.exception, that.exception);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ExceptionTableElement'

    /**
     * This class implements a code value.
     */
    public static final class CodeValue implements ToValue {

        /** Maximum stack size. */
        private final int maxStack;

        /** Maximum locals size. */
        private final int maxLocals;

        /** Instructions. */
        private final List<? extends Instruction> instructions;

        /** Exception table. */
        private final List<ExceptionTableElement> exceptionTable;

        /** Attributes. */
        private final List<? extends Attribute> attributes;

        /**
         * Constructs a code value.
         * @param ms maximum stack size - should be in 0..65535
         * @param ml maximum locals size - should be in 0..65535
         * @param c instructions - should not be <tt>null</tt>
         * @param e exception table - should not be <tt>null</tt>
         * @param a attribute - should not be <tt>null</tt>,
         *          should also be code attributes
         */
        public CodeValue(final int ms,
                         final int ml,
                         final List<? extends Instruction> c,
                         final List<ExceptionTableElement> e,
                         final List<? extends Attribute> a) {
            assert (ms >= 0) && (ms <= 65535) : "invalid ms";
            assert (ml >= 0) && (ml <= 65535) : "invalid ml";
            assert c != null : "null c";
            assert e != null : "null e";
            assert a != null : "null a";
            assert areCodeAttributes(a) : "invalid attribute";
            this.maxStack = ms;
            this.maxLocals = ml;
            this.instructions = new LinkedList<Instruction>(c);
            this.exceptionTable = new LinkedList<ExceptionTableElement>(e);
            this.attributes = new LinkedList<Attribute>(a);
        } // end constructor(int, int, List<? extends Instruction>, ...)

        /**
         * Returns the maximum stack size.
         * @return the maximum stack size
         */
        public int getMaxStack() {
            return this.maxStack;
        } // end method 'getMaxStack()'

        /**
         * Returns the maximum locals size.
         * @return the maximum locals size
         */
        public int getMaxLocals() {
            return this.maxLocals;
        } // end method 'getMaxLocals()'

        /**
         * Returns the instruction list.
         * @return the instruction list
         */
        public List<? extends Instruction> getInstructions() {
            return new LinkedList<Instruction>(this.instructions);
        } // end method 'getInstructions()'

        /**
         * Returns the exception table.
         * @return the exception table
         */
        public List<ExceptionTableElement> getExceptionTable() {
            return new LinkedList<ExceptionTableElement>(this.exceptionTable);
        } // end method 'getExceptionTable()'

        /**
         * Returns the attributes.
         * @return the attributes
         */
        public List<? extends Attribute> getAttributes() {
            return new LinkedList<Attribute>(this.attributes);
        } // end method 'getAttributes()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            final StringBuilder sb = new StringBuilder();
            sb.append("CodeValue(maxstack=");
            sb.append(this.maxStack);
            sb.append(",maxlocals=");
            sb.append(this.maxLocals);
            sb.append(",instructions=");
            sb.append(this.instructions.toString());
            sb.append(",exceptiontable=");
            sb.append(this.exceptionTable.toString());
            sb.append(",attributes=");
            sb.append(this.attributes.toString());
            sb.append(")");
            return sb.toString();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof CodeValue) {
                final CodeValue that = (CodeValue) obj;
                return this.maxStack == that.maxStack
                    && this.maxLocals == that.maxLocals
                    && this.instructions.equals(that.instructions)
                    && this.exceptionTable.equals(that.exceptionTable)
                    && this.attributes.equals(that.attributes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.maxStack
                + this.maxLocals
                + this.instructions.hashCode()
                + this.exceptionTable.hashCode()
                + this.attributes.hashCode();
        } // end method 'hashCode()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(5, 0);
            b.set(0, Value.createFromLong(this.maxStack));
            b.set(1, Value.createFromLong(this.maxLocals));
            b.set(2, API.encodeList(this.instructions));
            b.set(3, API.encodeList(this.exceptionTable));
            b.set(4, API.encodeList(this.attributes));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        static CodeValue fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            final List<Instruction> instrs = new LinkedList<Instruction>();
            for (Value i : API.decodeList(b.get(2))) {
                instrs.add(Instruction.fromValue(i));
            } // end for
            final List<ExceptionTableElement> exns = new LinkedList<ExceptionTableElement>();
            for (Value e : API.decodeList(b.get(3))) {
                exns.add(ExceptionTableElement.fromValue(e));
            } // end for
            final List<Attribute> attrs = new LinkedList<Attribute>();
            for (Value a : API.decodeList(b.get(4))) {
                attrs.add(Attribute.fromValue(a));
            } // end for
            return new CodeValue(b.get(0).asLong(),
                                 b.get(1).asLong(),
                                 instrs,
                                 exns,
                                 attrs);
        } // end method 'fromValue(Value)'

    } // end inner-class 'CodeValue'

    /**
     * This class represents verification type infos used in stack frames.
     */
    public static abstract class VerificationTypeInfo implements ToValue {

        /** The tag for 'type' elements. */
        private static final int TAG_TYPE = Hash.hashVariant("Class_or_interface").asLong();

        /** The tag for 'array' elements. */
        private static final int TAG_ARRAY = Hash.hashVariant("Array_type").asLong();

        /**
         * This class represents verification type infos with no parameter.
         */
        private static final class Simple extends VerificationTypeInfo {

            /** Type name. */
            private final String name;

            /** Type code. */
            private final int code;

            /**
             * Constructs a verification type from name and code.
             * @param n type name - should not be <tt>null</tt>
             * @param c type code
             */
            private Simple(final String n, final int c) {
                assert n != null : "null n";
                this.name = n;
                this.code = c;
            } // end constructor(String, int)

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                return Value.createFromLong(this.code);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return this.name;
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.code;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Simple) {
                    final Simple that = (Simple) obj;
                    return this.code == that.code;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Simple'

        /** Verification type info for "top". */
        public static final VerificationTypeInfo TopVariableInfo = new Simple("TopVariableInfo", 0);

        /** Verification type info for "integer". */
        public static final VerificationTypeInfo IntegerVariableInfo = new Simple("IntegerVariableInfo", 1);

        /** Verification type info for "float". */
        public static final VerificationTypeInfo FloatVariableInfo = new Simple("FloatVariableInfo", 2);

        /** Verification type info for "long". */
        public static final VerificationTypeInfo LongVariableInfo = new Simple("LongVariableInfo", 3);

        /** Verification type info for "double". */
        public static final VerificationTypeInfo DoubleVariableInfo = new Simple("DoubleVariableInfo", 4);

        /** Verification type info for "null". */
        public static final VerificationTypeInfo NullVariableInfo = new Simple("NullVariableInfo", 5);

        /** Verification type info for "uninitialized this". */
        public static final VerificationTypeInfo UninitializedThisVariableInfo = new Simple("UninitializedThisVariableInfo", 6);

        /** Verification type info for "object". */
        public static final class ObjectVariableInfo extends VerificationTypeInfo {

            /** Parameter. */
            private final Object param;

            /**
             * Constructs an instance of the "object" variable info.
             * @param t type name - should not be <tt>null</tt>
             */
            public ObjectVariableInfo(final Instruction.TypeName t) {
                assert t != null : "null t";
                this.param = t;
            } // end constructor(Instruction.TypeName)

            /**
             * Constructs an instance of the "object" variable info.
             * @param d descriptor - should not be <tt>null</tt>
             */
            public ObjectVariableInfo(final Descriptor.Array d) {
                assert d != null : "null d";
                this.param = d;
            } // end constructor(Descriptor.Array)

            /**
             * Returns the underlying type name.
             * @return the underlying type name if any, <tt>null</tt> otherwise
             */
            public Instruction.TypeName getTypeName() {
                return this.param instanceof Instruction.TypeName
                    ? (Instruction.TypeName) this.param
                    : null;
            } // end method 'getTypeName()'

            /**
             * Returns the underlying descriptor.
             * @return the underlying descriptor if any, <tt>null</tt> otherwise
             */
            public Descriptor.Array getDescriptor() {
                return this.param instanceof Descriptor.Array
                    ? (Descriptor.Array) this.param
                    : null;
            } // end method 'getDescriptor()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                if (this.param instanceof Instruction.TypeName) {
                    final Block b = Block.createBlock(1, 0);
                    b.set(0, Instruction.wrap("Class_or_interface", ((Instruction.TypeName) this.param).toValue()));
                    return Value.createFromBlock(b);
                } else {
                    final Block b = Block.createBlock(1, 0);
                    b.set(0, Instruction.wrap("Array", ((Instruction.TypeName) this.param).toValue()));
                    return Value.createFromBlock(b);
                } // end if/else
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "ObjectVariableInfo(" +  this.param.toString() + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.param.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof ObjectVariableInfo) {
                    final ObjectVariableInfo that = (ObjectVariableInfo) obj;
                    return this.param.equals(that.param);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'ObjectVariableInfo'

        /** Verification type info for "uninitialized". */
        public static final class UninitializedVariableInfo extends VerificationTypeInfo {

            /** Code offset of object creation. */
            private final int offset;

            /**
             * Constructs an instance of the "uninitialized" variable info.
             * @param o code offset of object creation - should be in 0..65535
             */
            public UninitializedVariableInfo(final int o) {
                assert (o >= 0) && (o <= 65535) : "invalid o";
                this.offset = o;
            } // end constructor(int)

            /**
             * Returns the code offset of object creation.
             * @return the code offset of object creation
             */
            public int getOffset() {
                return this.offset;
            } // end method 'getOffset()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(1, Value.createFromLong(this.offset));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "UninitializedVariableInfo(" + this.offset + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof UninitializedVariableInfo) {
                    final UninitializedVariableInfo that = (UninitializedVariableInfo) obj;
                    return this.offset == that.offset;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'UninitializedVariableInfo'

        private VerificationTypeInfo() {
        } // end empty constructor

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        @SuppressWarnings("fallthrough")
        static VerificationTypeInfo fromValue(final Value v) {
            assert v != null : "null v";
            if (v.isLong()) {
                switch (v.asLong()) {
                case 0: return TopVariableInfo;
                case 1: return IntegerVariableInfo;
                case 2: return FloatVariableInfo;
                case 3: return LongVariableInfo;
                case 4: return DoubleVariableInfo;
                case 5: return NullVariableInfo;
                case 6: return UninitializedThisVariableInfo;
                default:
                    assert false : "invalid tag";
                    return null;
                } // end switch
            } else {
                final Block b = v.asBlock();
                switch (b.getTag()) {
                case 0: if (b.get(0).asBlock().get(0).asLong() == TAG_TYPE) return new ObjectVariableInfo(Instruction.TypeName.fromValue(b.get(0).asBlock().get(1))); else if (b.get(0).asBlock().get(0).asLong() == TAG_ARRAY) return new ObjectVariableInfo((Descriptor.Array) Descriptor.fromValue(b.get(0).asBlock().get(1)));
                case 1: return new UninitializedVariableInfo((short) b.get(0).asBlock().get(1).asLong());
                default:
                    assert false : "invalid tag";
                    return null;
                } // end switch
            } // end if/else
        } // end method 'fromValue(Value)'

    } // end inner-class 'VerificationTypeInfo'

    /**
     * This class represents stack map frames.
     */
    public static abstract class StackMapFrame implements ToValue {

        /** Code offset. */
        protected final int offset;

        /**
         * Constructs an instance from a code offset.
         * @param o code offset - should be in 0..65535
         */
        private StackMapFrame(final int o) {
            assert (o >= 0) && (o <= 65535) : "invalid o";
            this.offset = o;
        } // end constructor(int)

        /**
         * Returns the code offset.
         * @return the code offset
         */
        public int getOffset() {
            return this.offset;
        } // end method 'getOffset()'

        /** Stack map frame for 'same'. */
        public static final class Same extends StackMapFrame {

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             */
            public Same(final int o) {
                super(o);
            } // end constructor(int)

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(0, Value.createFromLong(this.offset));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Same(" + this.offset + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Same) {
                    final Same that = (Same) obj;
                    return this.offset == that.offset;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Same'

        /** Stack map frame for 'same locals, 1 stack item'. */
        public static final class SameLocals1StackItem extends StackMapFrame {

            /** Type of stack item. */
            private final VerificationTypeInfo type;

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             * @param t type of stack item - should not be <tt>null</tt>
             */
            public SameLocals1StackItem(final int o, final VerificationTypeInfo t) {
                super(o);
                assert t != null : "null t";
                this.type = t;
            } // end constructor(int, VerificationTypeInfo)

            /**
             * Returns the type of the stack item.
             * @return the type of the stack item
             */
            public VerificationTypeInfo getType() {
                return this.type;
            } // end method 'getType()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(1, Value.createFromLong(this.offset), this.type.toValue());
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "SameLocals1StackItem(" + this.offset + ", " + this.type + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset + this.type.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof SameLocals1StackItem) {
                    final SameLocals1StackItem that = (SameLocals1StackItem) obj;
                    return (this.offset == that.offset)
                        && this.type.equals(that.type);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'SameLocals1StackItem'

        /** Stack map frame for 'chop1'. */
        public static final class Chop1 extends StackMapFrame {

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             */
            public Chop1(final int o) {
                super(o);
            } // end constructor(int)

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(2, Value.createFromLong(this.offset));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Chop1(" + this.offset + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Chop1) {
                    final Chop1 that = (Chop1) obj;
                    return this.offset == that.offset;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Chop1'

        /** Stack map frame for 'chop2'. */
        public static final class Chop2 extends StackMapFrame {

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             */
            public Chop2(final int o) {
                super(o);
            } // end constructor(int)

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(3, Value.createFromLong(this.offset));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Chop2(" + this.offset + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Chop2) {
                    final Chop2 that = (Chop2) obj;
                    return this.offset == that.offset;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Chop2'

        /** Stack map frame for 'chop3'. */
        public static final class Chop3 extends StackMapFrame {

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             */
            public Chop3(final int o) {
                super(o);
            } // end constructor(int)

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(4, Value.createFromLong(this.offset));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Chop3(" + this.offset + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset;
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Chop3) {
                    final Chop3 that = (Chop3) obj;
                    return this.offset == that.offset;
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Chop3'

        /** Stack map frame for 'append1'. */
        public static final class Append1 extends StackMapFrame {

            /** Type of stack item. */
            private final VerificationTypeInfo type;

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             * @param t type of stack item - should not be <tt>null</tt>
             */
            public Append1(final int o, final VerificationTypeInfo t) {
                super(o);
                assert t != null : "null t";
                this.type = t;
            } // end constructor(int, VerificationTypeInfo)

            /**
             * Returns the type of the stack item.
             * @return the type of the stack item
             */
            public VerificationTypeInfo getType() {
                return this.type;
            } // end method 'getType()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(5, Value.createFromLong(this.offset), this.type.toValue());
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Append1(" + this.offset + ", " + this.type + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset + this.type.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Append1) {
                    final Append1 that = (Append1) obj;
                    return (this.offset == that.offset)
                        && this.type.equals(that.type);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Append1'

        /** Stack map frame for 'append2'. */
        public static final class Append2 extends StackMapFrame {

            /** Type of first added stack item. */
            private final VerificationTypeInfo type1;

            /** Type of second added stack item. */
            private final VerificationTypeInfo type2;

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             * @param t1 type of stack item - should not be <tt>null</tt>
             * @param t2 type of stack item - should not be <tt>null</tt>
             */
            public Append2(final int o,
                           final VerificationTypeInfo t1,
                           final VerificationTypeInfo t2) {
                super(o);
                assert t1 != null : "null t1";
                assert t2 != null : "null t2";
                this.type1 = t1;
                this.type2 = t2;
            } // end constructor(int, VerificationTypeInfo, VerificationTypeInfo)

            /**
             * Returns the type of the first added stack item.
             * @return the type of the first added stack item
             */
            public VerificationTypeInfo getType1() {
                return this.type1;
            } // end method 'getType1()'

            /**
             * Returns the type of the second added stack item.
             * @return the type of the second added stack item
             */
            public VerificationTypeInfo getType2() {
                return this.type2;
            } // end method 'getType2()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(6, Value.createFromLong(this.offset), this.type1.toValue(), this.type2.toValue());
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Append2(" + this.offset + ", " + this.type1 + ", " + this.type2 + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset + this.type1.hashCode() + this.type2.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Append2) {
                    final Append2 that = (Append2) obj;
                    return (this.offset == that.offset)
                        && this.type1.equals(that.type1)
                        && this.type2.equals(that.type2);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Append2'

        /** Stack map frame for 'append3'. */
        public static final class Append3 extends StackMapFrame {

            /** Type of first added stack item. */
            private final VerificationTypeInfo type1;

            /** Type of second added stack item. */
            private final VerificationTypeInfo type2;

            /** Type of third added stack item. */
            private final VerificationTypeInfo type3;

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             * @param t1 type of stack item - should not be <tt>null</tt>
             * @param t2 type of stack item - should not be <tt>null</tt>
             * @param t3 type of stack item - should not be <tt>null</tt>
             */
            public Append3(final int o,
                           final VerificationTypeInfo t1,
                           final VerificationTypeInfo t2,
                           final VerificationTypeInfo t3) {
                super(o);
                assert t1 != null : "null t1";
                assert t2 != null : "null t2";
                assert t3 != null : "null t3";
                this.type1 = t1;
                this.type2 = t2;
                this.type3 = t3;
            } // end constructor(int, VerificationTypeInfo, VerificationTypeInfo, VerificationTypeInfo)

            /**
             * Returns the type of the first added stack item.
             * @return the type of the first added stack item
             */
            public VerificationTypeInfo getType1() {
                return this.type1;
            } // end method 'getType1()'

            /**
             * Returns the type of the second added stack item.
             * @return the type of the second added stack item
             */
            public VerificationTypeInfo getType2() {
                return this.type2;
            } // end method 'getType2()'

            /**
             * Returns the type of the third added stack item.
             * @return the type of the third added stack item
             */
            public VerificationTypeInfo getType3() {
                return this.type3;
            } // end method 'getType3()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(7, Value.createFromLong(this.offset), this.type1.toValue(), this.type2.toValue(), this.type3.toValue());
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Append3(" + this.offset + ", " + this.type1 + ", " + this.type2 + ", " + this.type3 + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset + this.type1.hashCode() + this.type2.hashCode() + this.type3.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Append3) {
                    final Append3 that = (Append3) obj;
                    return (this.offset == that.offset)
                        && this.type1.equals(that.type1)
                        && this.type2.equals(that.type2)
                        && this.type3.equals(that.type3);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Append3'

        /** Stack map frame for 'full'. */
        public static final class Full extends StackMapFrame {

            /** Type of first added stack item. */
            private final List<? extends VerificationTypeInfo> locals;

            /** Type of second added stack item. */
            private final List<? extends VerificationTypeInfo> operands;

            /**
             * Constructs an instance from a code offset.
             * @param o code offset - should be in 0..65535
             * @param locs types of locals - should not be <tt>null</tt>
             * @param ops types of stack operands - should not be <tt>null</tt>
             */
            public Full(final int o,
                        final List<? extends VerificationTypeInfo> locs,
                        final List<? extends VerificationTypeInfo> ops) {
                super(o);
                assert locs != null : "null locs";
                assert ops != null : "null ops";
                this.locals = new LinkedList<VerificationTypeInfo>(locs);
                this.operands = new LinkedList<VerificationTypeInfo>(ops);
            } // end constructor(int, List<? extends VerificationTypeInfo>, List<? extends VerificationTypeInfo>)

            /**
             * Returns the types of the locals.
             * @return the types of the locals
             */
            public List<? extends VerificationTypeInfo> getLocals() {
                return new LinkedList<VerificationTypeInfo>(this.locals);
            } // end method 'getLocals()'

            /**
             * Returns the types of the operands.
             * @return the types of the operands
             */
            public List<? extends VerificationTypeInfo> getOperands() {
                return new LinkedList<VerificationTypeInfo>(this.operands);
            } // end method 'getOperands()'

            /**
             * {@inheritDoc}
             */
            public Value toValue() {
                final Block b = Block.createBlock(8, Value.createFromLong(this.offset), API.encodeList(this.locals), API.encodeList(this.operands));
                return Value.createFromBlock(b);
            } // end method 'toValue()'

            /**
             * {@inheritDoc}
             */
            @Override
            public String toString() {
                return "Full(" + this.offset + ", " + this.locals + ", " + this.operands + ")";
            } // end method 'toString()'

            /**
             * {@inheritDoc}
             */
            @Override
            public int hashCode() {
                return this.offset + this.locals.hashCode() + this.operands.hashCode();
            } // end method 'hashCode()'

            /**
             * {@inheritDoc}
             */
            @Override
            public boolean equals(final Object obj) {
                if (obj instanceof Full) {
                    final Full that = (Full) obj;
                    return (this.offset == that.offset)
                        && this.locals.equals(that.locals)
                        && this.operands.equals(that.operands);
                } else {
                    return false;
                } // end if/else
            } // end method 'equals(Object)'

        } // end inner-class 'Full'

        /**
         * Constructs an instance from a value.
         * @param v value to construct instance from - should not be <tt>null</tt>
         * @return the corresponding instance
         */
        static StackMapFrame fromValue(final Value v) {
            assert v != null : "null v";
            final Block b = v.asBlock();
            switch (b.getTag()) {
            case 0: return new Same(b.get(0).asLong());
            case 1: return new SameLocals1StackItem(b.get(0).asLong(), VerificationTypeInfo.fromValue(b.get(1)));
            case 2: return new Chop1(b.get(0).asLong());
            case 3: return new Chop2(b.get(0).asLong());
            case 4: return new Chop3(b.get(0).asLong());
            case 5: return new Append1(b.get(0).asLong(), VerificationTypeInfo.fromValue(b.get(1)));
            case 6: return new Append2(b.get(0).asLong(), VerificationTypeInfo.fromValue(b.get(1)), VerificationTypeInfo.fromValue(b.get(2)));
            case 7: return new Append3(b.get(0).asLong(), VerificationTypeInfo.fromValue(b.get(1)), VerificationTypeInfo.fromValue(b.get(2)), VerificationTypeInfo.fromValue(b.get(3)));
            case 8:
                final List<VerificationTypeInfo> locals = new LinkedList<VerificationTypeInfo>();
                for (Value l : API.decodeList(b.get(1))) {
                    locals.add(VerificationTypeInfo.fromValue(l));
                } // end for
                final List<VerificationTypeInfo> operands = new LinkedList<VerificationTypeInfo>();
                for (Value o : API.decodeList(b.get(2))) {
                    operands.add(VerificationTypeInfo.fromValue(o));
                } // end for
                return new Full(b.get(0).asLong(), locals, operands);
            default:
                assert false : "invalid tag";
                return null;
            } // end switch
        } // end method 'fromValue(Value)'

    } // end inner-class 'StackMapFrame'

    /** Whether the attribute can be a field attribute. */
    private final boolean isField;

    /** Whether the attribute can be a method attribute. */
    private final boolean isMethod;

    /** Whether the attribute can be a class attribute. */
    private final boolean isClass;

    /** Whether the attribute can be a code attribute. */
    private final boolean isCode;

    /**
     * Constructs an attribute.
     * @param f whether the attribute can be a field attribute
     * @param m whether the attribute can be a method attribute
     * @param c whether the attribute can be a class attribute
     * @param d whether the attribute can be a code attribute
     */
    private Attribute(final boolean f,
                      final boolean m,
                      final boolean c,
                      final boolean d) {
        this.isField = f;
        this.isMethod = m;
        this.isClass = c;
        this.isCode = d;
    } // end constructor(boolean, boolean, boolean, boolean)

    /**
     * Tests whether the attribute can be a field attribute.
     * @return <tt>true</tt> if the attribute can be a field attribute,
     *         <tt>false</tt> otherwise
     */
    public final boolean isFieldAttribute() {
        return this.isField;
    } // end method 'isFieldAttribute()'

    /**
     * Tests whether the attribute can be a method attribute.
     * @return <tt>true</tt> if the attribute can be a method attribute,
     *         <tt>false</tt> otherwise
     */
    public final boolean isMethodAttribute() {
        return this.isMethod;
    } // end method 'isMethodAttribute()'

    /**
     * Tests whether the attribute can be a class attribute.
     * @return <tt>true</tt> if the attribute can be a class attribute,
     *         <tt>false</tt> otherwise
     */
    public final boolean isClassAttribute() {
        return this.isClass;
    } // end method 'isClassAttribute()'

    /**
     * Tests whether the attribute can be a code attribute.
     * @return <tt>true</tt> if the attribute can be a code attribute,
     *         <tt>false</tt> otherwise
     */
    public final boolean isCodeAttribute() {
        return this.isCode;
    } // end method 'isCodeAttribute()'

    /**
     * {@inheritDoc}
     */
    @Override
    public final String toString() {
        final Class c = this.getClass();
        if (c.equals(Deprecated.class)) {
            return "Deprecated";
        } else if (c.equals(Synthetic.class)) {
            return "Synthetic";
        } else {
            return c.getSimpleName() + "(-)";
        } // end if/elsif/else
    } // end method 'toString()'

    /**
     * {@inheritDoc}
     */
    @Override
    public final int hashCode() {
        return this.getClass().hashCode();
    } // end method 'toString()'

    /**
     * Tests whether a list contains only field attributes.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only field attributes,
     *         <tt>false</tt> otherwise
     */
    public static boolean areFieldAttributes(final List<? extends Attribute> l) {
        assert l != null : "null l";
        for (Attribute a : l) {
            if (!a.isFieldAttribute()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areFieldAttributes(List<? extends Attribute>)'

    /**
     * Tests whether a list contains only method attributes.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only method attributes,
     *         <tt>false</tt> otherwise
     */
    public static boolean areMethodAttributes(final List<? extends Attribute> l) {
        assert l != null : "null l";
        for (Attribute a : l) {
            if (!a.isMethodAttribute()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areMethodAttributes(List<? extends Attribute>)'

    /**
     * Tests whether a list contains only class attributes.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only class attributes,
     *         <tt>false</tt> otherwise
     */
    public static boolean areClassAttributes(final List<? extends Attribute> l) {
        assert l != null : "null l";
        for (Attribute a : l) {
            if (!a.isClassAttribute()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areClassAttributes(List<? extends Attribute>)'

    /**
     * Tests whether a list contains only code attributes.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only code attributes,
     *         <tt>false</tt> otherwise
     */
    public static boolean areCodeAttributes(final List<? extends Attribute> l) {
        assert l != null : "null l";
        for (Attribute a : l) {
            if (!a.isCodeAttribute()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areCodeAttributes(List<? extends Attribute>)'

    /**
     * This class implements an 'AnnotationDefault' attribute.
     */
    public static final class AnnotationDefault extends Attribute {

        /** Attribute value. */
        private final Annotation.ElementValue value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public AnnotationDefault(final Annotation.ElementValue v) {
            super(false, true, false, false);
            assert v != null : "null v";
            this.value = v;
        } // end constructor(Annotation.ElementValue)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public Annotation.ElementValue getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_ANNOTATION_DEFAULT,
                                              this.value.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof AnnotationDefault) {
                final AnnotationDefault that = (AnnotationDefault) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'AnnotationDefault'

    /**
     * This class implements a 'Code' attribute.
     */
    public static final class Code extends Attribute {

        /** Attribute value. */
        private final CodeValue value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public Code(final CodeValue v) {
            super(false, true, false, false);
            assert v != null : "null v";
            this.value = v;
        } // end constructor(CodeValue)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public CodeValue getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_CODE,
                                              this.value.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Code) {
                final Code that = (Code) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Code'

    /**
     * This class implements a 'ConstantValue' attribute.
     */
    public static final class ConstantValue extends Attribute {

        /** Attribute value. */
        private final ConstantVal value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public ConstantValue(final ConstantVal v) {
            super(true, false, false, false);
            assert v != null : "null v";
            this.value = v;
        } // end constructor(ConstantVal)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public ConstantVal getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_CONSTANT_VALUE,
                                              this.value.toValue());
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof ConstantValue) {
                final ConstantValue that = (ConstantValue) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'ConstantValue'

    /**
     * This class implements a 'Deprecated' attribute.
     */
    public static final class Deprecated extends Attribute {

        /**
         * Constructs an attribute.
         */
        public Deprecated() {
            super(true, true, true, false);
        } // end empty constructor

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            return HASH_DEPRECATED;
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Deprecated) {
                return true;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Deprecated'

    /**
     * This class implements an 'EnclosingMethod' attribute.
     */
    public static final class EnclosingMethod extends Attribute {

        /** Class name. */
        private final String className;

        /** Method name. */
        private final String methodName;

        /** Method return type descriptor. */
        private final Descriptor returnType;

        /** Method parameter type descriptors. */
        private final List<? extends Descriptor> parameterTypes;

        /**
         * Constructs an attribute.
         * @param cn class name - should not be <tt>null</tt>
         * @param mn method name - should not be <tt>null</tt>
         * @param mr method return type descriptor - should not be <tt>null</tt>
         * @param mp method parameter type descriptors - should not be <tt>null</tt>,
         *           should also be non-void descriptors
         * @throws BaristaException if one of the descriptors is not a valid descriptor
         * @throws BaristaException if one of the attributes is not a valid attribute
         */
        public EnclosingMethod(final String cn,
                               final String mn,
                               final Descriptor mr,
                               final List<? extends Descriptor> mp)
            throws BaristaException {
            super(false, false, true, false);
            assert cn != null : "null cn";
            assert mn != null : "null mn";
            assert mr != null : "null mr";
            assert mp != null : "null mp";
            if (!mr.isFieldDescriptor()) {
                throw new BaristaException("invalid descriptor for return type");
            } // end if
            if (!Descriptor.areFieldDescriptors(mp)) {
                throw new BaristaException("invalid descriptor for parameter type");
            } // end fi
            this.className = cn;
            this.methodName = mn;
            this.returnType = mr;
            this.parameterTypes = new LinkedList<Descriptor>(mp);
        } // end constructor(String, String, Descriptor, List<? extends Descriptor>)

        /**
         * Constructs an attribute.
         * @param cn class name - should not be <tt>null</tt>
         */
        public EnclosingMethod(final String cn) {
            super(false, false, true, false);
            assert cn != null : "null cn";
            this.className = cn;
            this.methodName = null;
            this.returnType = null;
            this.parameterTypes = null;
        } // end constructor(String)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getClassName() {
            return this.className;
        } // end method 'getClassName()'

        /**
         * Returns the method name.
         * @return the method name
         */
        public String getMethodName() {
            return this.methodName;
        } // end method 'getMethodName()'

        /**
         * Returns the method return type.
         * @return the method return type
         */
        public Descriptor getReturnType() {
            return this.returnType;
        } // end method 'getReturnType()'

        /**
         * Returns the method parameter types.
         * @return the method parameter types
         */
        public List<? extends Descriptor> getParameterTypes() {
            return new LinkedList<Descriptor>(this.parameterTypes);
        } // end method 'getParameterTypes()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Value o;
            if (this.methodName == null) {
                o = Value.ZERO;
            } else {
                final Block desc = Block.createBlock(0,
                                                     API.encodeList(this.parameterTypes),
                                                     this.returnType.toValue());
                final Block m = Block.createBlock(0,
                                                  API.createMethodName(this.methodName),
                                                  Value.createFromBlock(desc));
                o = Value.createFromBlock(m);
            } // end if/else
            final Block c = Block.createBlock(0,
                                              API.createClassName(this.className),
                                              o);
            final Block b = Block.createBlock(0,
                                              HASH_ENCLOSING_METHOD,
                                              Value.createFromBlock(c));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof EnclosingMethod) {
                final EnclosingMethod that = (EnclosingMethod) obj;
                return this.className.equals(that.className)
                    && API.nullOrEquals(this.methodName, that.methodName)
                    && API.nullOrEquals(this.returnType, that.returnType)
                    && API.nullOrEquals(this.parameterTypes, that.parameterTypes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'EnclosingMethod'

    /**
     * This class implements an 'Exceptions' attribute.
     */
    public static final class Exceptions extends Attribute {

        /** Attribute value. */
        private final List<String> value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public Exceptions(final List<String> v) {
            super(false, true, false, false);
            assert v != null : "null v";
            this.value = new LinkedList<String>(v);
        } // end constructor(List<String>)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public List<String> getValue() {
            return new LinkedList<String>(this.value);
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final List<Value> l = new LinkedList<Value>();
            for (String s : this.value) {
                l.add(API.createClassName(s));
            } // end value
            final Block b = Block.createBlock(0,
                                              HASH_EXCEPTIONS,
                                              API.encodeValueList(l));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Exceptions) {
                final Exceptions that = (Exceptions) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Exceptions'

    /**
     * This class implements an 'InnerClasses' attribute.
     */
    public static final class InnerClasses extends Attribute {

        /** Inner-class names. */
        private final List<String> inners;

        /** Outer-class names. */
        private final List<String> outers;

        /** Class names. */
        private final List<String> names;

        /** Class flags. */
        private final List< List<AccessFlag> > flags;

        /**
         * Constructs an attribute. <br>
         * <b>all passed lists should have the same size.</b>
         * @param i inner-class names - should not be <tt>null</tt>
         * @param o outer-class names - should not be <tt>null</tt>
         * @param n names - should not be <tt>null</tt>
         * @param f class flags - should not be <tt>null</tt>,
         *          should also be inner-class flags
         * @throws BaristaException if one of the flags is not a valid flag
         */
        public InnerClasses(final List<String> i,
                            final List<String> o,
                            final List<String> n,
                            final List< List<AccessFlag> > f)
            throws BaristaException {
            super(false, false, true, false);
            assert i != null : "null i";
            assert o != null : "null o";
            assert n != null : "null n";
            assert f != null : "null f";
            for (List<AccessFlag> l : f) {
                if (!AccessFlag.areInnerClassFlags(l)) {
                    throw new BaristaException("invalid flag");
                } // endif
            } // end for
            assert i.size() == o.size() 
                && o.size() == n.size()
                && n.size() == f.size() : "parameters have different sizes";
            this.inners = new LinkedList<String>(i);
            this.outers = new LinkedList<String>(o);
            this.names = new LinkedList<String>(n);
            this.flags = new LinkedList< List<AccessFlag> >();
            for (List<AccessFlag> l : f) {
                this.flags.add(new LinkedList<AccessFlag>(l));
            } // end for
        } // end constructor(List<String>, List<String>, ...)

        /**
         * Returns the inner class names.
         * @return the inner class names
         */
        public List<String> getInners() {
            return new LinkedList<String>(this.inners);
        } // end method 'getInners()'

        /**
         * Returns the outer class names.
         * @return the outer class names
         */
        public List<String> getOuters() {
            return new LinkedList<String>(this.outers);
        } // end method 'getOuters()'

        /**
         * Returns the class names.
         * @return the class names
         */
        public List<String> getNames() {
            return new LinkedList<String>(this.names);
        } // end method 'getNames()'

        /**
         * Returns the class flags.
         * @return the class flags
         */
        public List< List<AccessFlag> > getFlags() {
            final List< List<AccessFlag> > res = new LinkedList< List<AccessFlag> >();
            for (List<AccessFlag> l : this.flags) {
                res.add(new LinkedList<AccessFlag>(l));
            } // end for
            return res;
        } // end method 'getFlags()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final List<Value> l = new LinkedList<Value>();
            final Iterator<String> itInn = this.inners.iterator();
            final Iterator<String> itOut = this.outers.iterator();
            final Iterator<String> itNames = this.names.iterator();
            final Iterator< List<AccessFlag> > itFlags = this.flags.iterator();
            while (itInn.hasNext() && itOut.hasNext()
                   && itNames.hasNext() && itFlags.hasNext()) {
                final String i = itInn.next();
                final String o = itOut.next();
                final String n = itNames.next();
                final List<AccessFlag> f = itFlags.next();
                final Value bi;
                if (i == null) {
                    bi = Value.ZERO;
                } else {
                    bi = Value.createFromBlock(Block.createBlock(0, API.createClassName(i)));
                } // end if/else
                final Value bo;
                if (o == null) {
                    bo = Value.ZERO;
                } else {
                    bo = Value.createFromBlock(Block.createBlock(0, API.createClassName(o)));
                } // end if/else
                final Value bn;
                if (n == null) {
                    bn = Value.ZERO;
                } else {
                    bn = Value.createFromBlock(Block.createBlock(0, API.createUTF8(n)));
                } // end if/else
                l.add(Value.createFromBlock(Block.createBlock(0, bi, bo, bn, API.encodeList(f))));
            } // end while
            final Block b = Block.createBlock(0,
                                              HASH_INNER_CLASSES,
                                              API.encodeValueList(l));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof InnerClasses) {
                final InnerClasses that = (InnerClasses) obj;
                return this.inners.equals(that.inners)
                    && this.outers.equals(that.outers)
                    && this.names.equals(that.names)
                    && this.flags.equals(that.flags);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'InnerClasses'

    /**
     * This class implements a 'LineNumberTable' attribute.
     */
    public static final class LineNumberTable extends Attribute {

        /** Offets. */
        private final List<Integer> offsets;

        /** Lines. */
        private final List<Integer> lines;

        /**
         * Constructs an attribute.
         * @param ofs offets - should not be <tt>null</tt>,
         *            each value should be in 0..65535
         * @param ln lines - should not be <tt>null</tt>,
         *           each value should be in 0..65535
         */
        public LineNumberTable(final List<Integer> ofs,
                               final List<Integer> ln) {
            super(false, false, false, true);
            assert ofs != null : "null ofs";
            assert ln != null : "null ln";
            assert ofs.size() == ln.size() : "ofs and ln have different sizes";
            for (int o : ofs) {
                assert (o >= 0) && (o <= 65535) : "invalid offset value";
            } // end for
            for (int l : ln) {
                assert (l >= 0) && (l <= 65535) : "invalid line value";
            } // end for
            this.offsets = new LinkedList<Integer>(ofs);
            this.lines = new LinkedList<Integer>(ln);
        } // end constructor(List<Integer>, List<Integer>)

        /**
         * Returns the offsets.
         * @return the offsets
         */
        public List<Integer> getOffsets() {
            return new LinkedList<Integer>(this.offsets);
        } // end method 'getOffsets()'

        /**
         * Returns the lines.
         * @return the lines
         */
        public List<Integer> getLines() {
            return new LinkedList<Integer>(this.lines);
        } // end method 'getLines()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final List<Value> l = new LinkedList<Value>();
            final Iterator<Integer> itOffsets = this.offsets.iterator();
            final Iterator<Integer> itLines = this.lines.iterator();
            while (itOffsets.hasNext() && itLines.hasNext()) {
                final Block e = Block.createBlock(0,
                                                  Value.createFromLong(itOffsets.next()),
                                                  Value.createFromLong(itLines.next()));
                l.add(Value.createFromBlock(e));
            } // end while
            final Block b = Block.createBlock(0,
                                              HASH_LINE_NUMBER_TABLE,
                                              API.encodeValueList(l));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof LineNumberTable) {
                final LineNumberTable that = (LineNumberTable) obj;
                return this.offsets.equals(that.offsets)
                    && this.lines.equals(that.lines);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LineNumberTable'

    /**
     * This class implements a 'LocalVariableTable' attribute.
     */
    public static final class LocalVariableTable extends Attribute {

        /** Starting offets. */
        private final List<Integer> starts;

        /** Ending offets. */
        private final List<Integer> ends;

        /** Variable names. */
        private final List<String> names;

        /** Variable descriptors. */
        private final List<? extends Descriptor> descriptors;

        /** Variable indexes. */
        private final List<Integer> indexes;

        /**
         * Constructs an attribute. <br>
         * <b>all passed lists should have the same size.</b>
         * @param s starting offsets - should not be <tt>null</t>,
         *          each value should be in 0..65535
         * @param e ending offsets - should not be <tt>null</t>,
         *          each value should be in 0..65535
         * @param n variable names - should not be <tt>null</t>
         * @param d descriptors - should not be <tt>null</t>,
         *          should also be a non-void descriptor
         * @param i indexes - should not be <tt>null</t>,
         *          each value should be in 0..65535
         * @throws BaristaException if one of the descriptors is not a valid descriptor
         */
        public LocalVariableTable(final List<Integer> s,
                                  final List<Integer> e,
                                  final List<String> n,
                                  final List<? extends Descriptor> d,
                                  final List<Integer> i)
            throws BaristaException {
            super(false, false, false, true);
            assert s != null : "null s";
            assert e != null : "null e";
            assert n != null : "null n";
            assert d != null : "null d";
            assert i != null : "null i";
            if (!Descriptor.areFieldDescriptors(d)) {
                throw new BaristaException("invalid descriptor");
            } // end if
            assert s.size() == e.size() 
                && e.size() == n.size()
                && n.size() == d.size()
                && d.size() == i.size() : "parameters have different sizes";
            for (int x : s) {
                assert (x >= 0) && (x <= 65535) : "invalid offset value";
            } // end for
            for (int x : e) {
                assert (x >= 0) && (x <= 65535) : "invalid offset value";
            } // end for
            for (int x : i) {
                assert (x >= 0) && (x <= 65535) : "invalid index value";
            } // end for
            this.starts = new LinkedList<Integer>(s);
            this.ends = new LinkedList<Integer>(e);
            this.names = new LinkedList<String>(n);
            this.descriptors = new LinkedList<Descriptor>(d);
            this.indexes = new LinkedList<Integer>(i);
        } // end constructor(List<Integer>, List<Integer>, ...)

        /**
         * Returns the list of starting offsets.
         * @return the list of starting offsets
         */
        public List<Integer> getStartingOffsets() {
            return new LinkedList<Integer>(this.starts);
        } // end method 'getStartingOffsets()'

        /**
         * Returns the list of ending offsets.
         * @return the list of ending offsets
         */
        public List<Integer> getEndingOffsets() {
            return new LinkedList<Integer>(this.ends);
        } // end method 'getEndingOffsets()'

        /**
         * Returns the list of names.
         * @return the list of names
         */
        public List<String> getNames() {
            return new LinkedList<String>(this.names);
        } // end method 'getNames()'

        /**
         * Returns the list of descriptors.
         * @return the list of descriptors
         */
        public List<? extends Descriptor> getDescriptors() {
            return new LinkedList<Descriptor>(this.descriptors);
        } // end method 'getDescriptors()'

        /**
         * Returns the list of indexes.
         * @return the list of indexes
         */
        public List<Integer> getIndexes() {
            return new LinkedList<Integer>(this.indexes);
        } // end method 'getIndexes()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final List<Value> l = new LinkedList<Value>();
            final Iterator<Integer> itStarts = this.starts.iterator();
            final Iterator<Integer> itEnds = this.ends.iterator();
            final Iterator<String> itNames = this.names.iterator();
            final Iterator<? extends Descriptor> itDescs = this.descriptors.iterator();
            final Iterator<Integer> itInd = this.indexes.iterator();
            while (itStarts.hasNext() && itEnds.hasNext() && itNames.hasNext()
                   && itDescs.hasNext() && itInd.hasNext()) {
                final Block e = Block.createBlock(5, 0);
                e.set(0, Value.createFromLong(itStarts.next()));
                e.set(1, Value.createFromLong(itEnds.next()));
                e.set(2, API.createUTF8(itNames.next()));
                e.set(3, itDescs.next().toValue());
                e.set(4, Value.createFromLong(itInd.next()));
                l.add(Value.createFromBlock(e));
            } // end while
            final Block b = Block.createBlock(0,
                                              HASH_LOCAL_VARIABLE_TABLE,
                                              API.encodeValueList(l));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof LocalVariableTable) {
                final LocalVariableTable that = (LocalVariableTable) obj;
                return this.starts.equals(that.starts)
                    && this.ends.equals(that.ends)
                    && this.names.equals(that.names)
                    && this.descriptors.equals(that.descriptors)
                    && this.indexes.equals(that.indexes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LocalVariableTable'

    /**
     * This class implements a 'LocalVariableTypeTable' attribute.
     */
    public static final class LocalVariableTypeTable extends Attribute {

        /** Starting offets. */
        private final List<Integer> starts;

        /** Ending offets. */
        private final List<Integer> ends;

        /** Variable names. */
        private final List<String> names;

        /** Variable type signatures. */
        private final List<String> signatures;

        /** Variable indexes. */
        private final List<Integer> indexes;

        /**
         * Constructs an attribute. <br>
         * <b>all passed lists should have the same size.</b>
         * @param s starting offsets - should not be <tt>null</t>,
         *          each value should be in 0..65535
         * @param e ending offsets - should not be <tt>null</t>,
         *          each value should be in 0..65535
         * @param n variable names - should not be <tt>null</t>
         * @param t type signatures - should not be <tt>null</t>
         * @param i indexes - should not be <tt>null</t>,
         *          each value should be in 0..65535
         */
        public LocalVariableTypeTable(final List<Integer> s,
                                      final List<Integer> e,
                                      final List<String> n,
                                      final List<String> t,
                                      final List<Integer> i) {
            super(false, false, false, true);
            assert s != null : "null s";
            assert e != null : "null e";
            assert n != null : "null n";
            assert t != null : "null t";
            assert i != null : "null i";
            assert s.size() == e.size() 
                && e.size() == n.size()
                && n.size() == t.size()
                && t.size() == i.size() : "parameters have different sizes";
            for (int x : s) {
                assert (x >= 0) && (x <= 65535) : "invalid offset value";
            } // end for
            for (int x : e) {
                assert (x >= 0) && (x <= 65535) : "invalid offset value";
            } // end for
            for (int x : i) {
                assert (x >= 0) && (x <= 65535) : "invalid index value";
            } // end for
            this.starts = new LinkedList<Integer>(s);
            this.ends = new LinkedList<Integer>(e);
            this.names = new LinkedList<String>(n);
            this.signatures = new LinkedList<String>(t);
            this.indexes = new LinkedList<Integer>(i);
        } // end constructor(List<Integer>, List<Integer>, ...)

        /**
         * Returns the list of starting offsets.
         * @return the list of starting offsets
         */
        public List<Integer> getStartingOffsets() {
            return new LinkedList<Integer>(this.starts);
        } // end method 'getStartingOffsets()'

        /**
         * Returns the list of ending offsets.
         * @return the list of ending offsets
         */
        public List<Integer> getEndingOffsets() {
            return new LinkedList<Integer>(this.ends);
        } // end method 'getEndingOffsets()'

        /**
         * Returns the list of names.
         * @return the list of names
         */
        public List<String> getNames() {
            return new LinkedList<String>(this.names);
        } // end method 'getNames()'

        /**
         * Returns the list of type signatures.
         * @return the list of type signatures
         */
        public List<String> getSignatures() {
            return new LinkedList<String>(this.signatures);
        } // end method 'getSignatures()'

        /**
         * Returns the list of indexes.
         * @return the list of indexes
         */
        public List<Integer> getIndexes() {
            return new LinkedList<Integer>(this.indexes);
        } // end method 'getIndexes()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final List<Value> l = new LinkedList<Value>();
            final Iterator<Integer> itStarts = this.starts.iterator();
            final Iterator<Integer> itEnds = this.ends.iterator();
            final Iterator<String> itNames = this.names.iterator();
            final Iterator<String> itSigns = this.signatures.iterator();
            final Iterator<Integer> itInd = this.indexes.iterator();
            while (itStarts.hasNext() && itEnds.hasNext() && itNames.hasNext()
                   && itSigns.hasNext() && itInd.hasNext()) {
                final Block e = Block.createBlock(5, 0);
                e.set(0, Value.createFromLong(itStarts.next()));
                e.set(1, Value.createFromLong(itEnds.next()));
                e.set(2, API.createUTF8(itNames.next()));
                e.set(3, API.createFieldSignature(itSigns.next()));
                e.set(4, Value.createFromLong(itInd.next()));
                l.add(Value.createFromBlock(e));
            } // end while
            final Block b = Block.createBlock(0,
                                              HASH_LOCAL_VARIABLE_TYPE_TABLE,
                                              API.encodeValueList(l));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof LocalVariableTypeTable) {
                final LocalVariableTypeTable that = (LocalVariableTypeTable) obj;
                return this.starts.equals(that.starts)
                    && this.ends.equals(that.ends)
                    && this.names.equals(that.names)
                    && this.signatures.equals(that.signatures)
                    && this.indexes.equals(that.indexes);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'LocalVariableTypeTable'

    /**
     * This class implements a 'RuntimeInvisibleAnnotations' attribute.
     */
    public static final class RuntimeInvisibleAnnotations extends Attribute {

        /** Attribute value. */
        private final List<Annotation> value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public RuntimeInvisibleAnnotations(final List<Annotation> v) {
            super(true, true, true, false);
            assert v != null : "null v";
            this.value = new LinkedList<Annotation>(v);
        } // end constructor(List<Annotation>)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public List<Annotation> getValue() {
            return new LinkedList<Annotation>(this.value);
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_RUNTIME_INVISIBLE_ANNOTATIONS,
                                              API.encodeList(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof RuntimeInvisibleAnnotations) {
                final RuntimeInvisibleAnnotations that = (RuntimeInvisibleAnnotations) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'RuntimeInvisibleAnnotations'

    /**
     * This class implements a 'RuntimeInvisibleParameterAnnotations' attribute.
     */
    public static final class RuntimeInvisibleParameterAnnotations extends Attribute {

        /** Attribute value. */
        private final List< List<Annotation> > value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public RuntimeInvisibleParameterAnnotations(final List< List<Annotation> > v) {
            super(false, true, false, false);
            assert v != null : "null v";
            this.value = new LinkedList< List<Annotation> >();
            for (List<Annotation> a : v) {
                this.value.add(new LinkedList<Annotation>(a));
            } // end for
        } // end constructor(List< List<Annotation> >)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public List< List<Annotation> > getValue() {
            final List< List<Annotation> > res = new LinkedList< List<Annotation> >();
            for (List<Annotation> l : this.value) {
                res.add(new LinkedList<Annotation>(l));
            } // end for
            return res;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_RUNTIME_INVISIBLE_PARAMETER_ANNOTATIONS,
                                              API.encodeListList(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof RuntimeInvisibleParameterAnnotations) {
                final RuntimeInvisibleParameterAnnotations that = (RuntimeInvisibleParameterAnnotations) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'RuntimeInvisibleParameterAnnotations'

    /**
     * This class implements a 'RuntimeVisibleAnnotations' attribute.
     */
    public static final class RuntimeVisibleAnnotations extends Attribute {

        /** Attribute value. */
        private final List<Annotation> value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public RuntimeVisibleAnnotations(final List<Annotation> v) {
            super(true, true, true, false);
            assert v != null : "null v";
            this.value = new LinkedList<Annotation>(v);
        } // end constructor(List<Annotation>)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public List<Annotation> getValue() {
            return new LinkedList<Annotation>(this.value);
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_RUNTIME_VISIBLE_ANNOTATIONS,
                                              API.encodeList(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof RuntimeVisibleAnnotations) {
                final RuntimeVisibleAnnotations that = (RuntimeVisibleAnnotations) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'RuntimeVisibleAnnotations'

    /**
     * This class implements a 'RuntimeVisibleParameterAnnotations' attribute.
     */
    public static final class RuntimeVisibleParameterAnnotations extends Attribute {

        /** Attribute value. */
        private final List< List<Annotation> > value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public RuntimeVisibleParameterAnnotations(final List< List<Annotation> > v) {
            super(false, true, false, false);
            assert v != null : "null v";
            this.value = new LinkedList< List<Annotation> >();
            for (List<Annotation> a : v) {
                this.value.add(new LinkedList<Annotation>(a));
            } // end for
        } // end constructor(List< List<Annotation> >)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public List< List<Annotation> > getValue() {
            final List< List<Annotation> > res = new LinkedList< List<Annotation> >();
            for (List<Annotation> l : this.value) {
                res.add(new LinkedList<Annotation>(l));
            } // end for
            return res;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_RUNTIME_VISIBLE_PARAMETER_ANNOTATIONS,
                                              API.encodeListList(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof RuntimeVisibleParameterAnnotations) {
                final RuntimeVisibleParameterAnnotations that = (RuntimeVisibleParameterAnnotations) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'RuntimeVisibleParameterAnnotations'

    /**
     * This class implements a 'Signature' attribute.
     */
    public static final class Signature extends Attribute {

        /** Hash value for 'Class' attribute. */
        private static final Value HASH_CLASS =
            Hash.hashVariant("Class");

        /** Hash value for 'Method' attribute. */
        private static final Value HASH_METHOD =
            Hash.hashVariant("Method");

        /** Hash value for 'Field' attribute. */
        private static final Value HASH_FIELD =
            Hash.hashVariant("Field");

        /** Kinds of signatures. */
        public static enum Kind { CLASS, METHOD, FIELD };

        /** Attribute kind. */
        private final Kind kind;

        /** Attribute value. */
        private final String value;

        /**
         * Constructs an attribute.
         * @param k kind of signature - should not be <tt>null</tt>
         * @param v attribute value - should not be <tt>null</tt>
         */
        public Signature(final Kind k, final String v) {
            super(k.equals(Kind.FIELD), k.equals(Kind.METHOD), k.equals(Kind.CLASS), false);
            assert k != null : "null k";
            assert v != null : "null v";
            this.kind = k;
            this.value = v;
        } // end constructor(String)

        /**
         * Returns the attribute kind.
         * @return the attribute kind
         */
        public Kind getKind() {
            return this.kind;
        } // end method 'getKind()'

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public String getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block v;
            switch (this.kind) {
            case CLASS:
                v = Block.createBlock(0,
                                      HASH_CLASS,
                                      API.createClassSignature(this.value));
                break;
            case METHOD:
                v = Block.createBlock(0,
                                      HASH_METHOD,
                                      API.createMethodSignature(this.value));
                break;
            case FIELD:
                v = Block.createBlock(0,
                                      HASH_FIELD,
                                      API.createFieldSignature(this.value));
                break;
            default:
                // should not be reached
                v = Block.createBlock(0, Value.ZERO, Value.ZERO);
                break;
            } // end switch
            final Block b = Block.createBlock(0,
                                              HASH_SIGNATURE,
                                              Value.createFromBlock(v));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Signature) {
                final Signature that = (Signature) obj;
                return this.kind.equals(that.kind)
                    && this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Signature'

    /**
     * This class implements a 'SourceDebugExtension' attribute.
     */
    public static final class SourceDebugExtension extends Attribute {

        /** Attribute value. */
        private final String value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public SourceDebugExtension(final String v) {
            super(false, false, true, false);
            assert v != null : "null v";
            this.value = v;
        } // end constructor(String)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public String getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_SOURCE_DEBUG_EXTENSION,
                                              API.createUTF8(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof SourceDebugExtension) {
                final SourceDebugExtension that = (SourceDebugExtension) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'SourceDebugExtension'

    /**
     * This class implements a 'SourceFile' attribute.
     */
    public static final class SourceFile extends Attribute {

        /** Attribute value. */
        private final String value;

        /**
         * Constructs an attribute.
         * @param v attribute value - should not be <tt>null</tt>
         */
        public SourceFile(final String v) {
            super(false, false, true, false);
            assert v != null : "null v";
            this.value = v;
        } // end constructor(String)

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public String getValue() {
            return this.value;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block b = Block.createBlock(0,
                                              HASH_SOURCE_FILE,
                                              API.createUTF8(this.value));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof SourceFile) {
                final SourceFile that = (SourceFile) obj;
                return this.value.equals(that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'SourceFile'

    /**
     * This class implements a 'Synthetic' attribute.
     */
    public static final class Synthetic extends Attribute {

        /**
         * Constructs an attribute.
         */
        public Synthetic() {
            super(true, true, true, false);
        } // end empty constructor

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            return HASH_SYNTHETIC;
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Synthetic) {
                return true;
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Synthetic'

    /**
     * This class implements a 'StackMapTable' attribute.
     */
    public static final class StackMapTable extends Attribute {

        /** Embedded frames. */
        private final List<? extends StackMapFrame> frames;

        /**
         * Constructs an attribute.
         * @param f stack frames - should not be <tt>null</tt>
         */
        public StackMapTable(final List<? extends StackMapFrame> f) {
            super(false, false, false, true);
            assert f != null : "null f";
            this.frames = new LinkedList<StackMapFrame>(f);
        } // end constructor(List<? extends StackMapFrame>)

        /**
         * Returns the list of embedded frames.
         * @return the list of embedded frames
         */
        public List<? extends StackMapFrame> getFrames() {
            return new LinkedList<StackMapFrame>(this.frames);
        } // end method 'getFrames()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block c = Block.createBlock(0, API.encodeList(this.frames));
            final Block b = Block.createBlock(0,
                                              HASH_STACK_MAP_TABLE,
                                              Value.createFromBlock(c));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof StackMapTable) {
                final StackMapTable that = (StackMapTable) obj;
                return this.frames.equals(that.frames);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'StackMapTable'

    /**
     * This class implements an 'Unknown' attribute.
     */
    public static final class Unknown extends Attribute {

        /** Attribute name. */
        private final String name;

        /** Attribute value. */
        private final byte[] value;

        /**
         * Constructs an attribute.
         * @param n attribute name - should not be <tt>null</tt>
         * @param v attribute value - should not be <tt>null</tt>
         */
        public Unknown(final String n, final byte[] v) {
            super(true, true, true, true);
            assert n != null : "null n";
            assert v != null : "null v";
            this.name = n;
            final int len = v.length;
            this.value = new byte[len];
            System.arraycopy(v, 0, this.value, 0, len);
        } // end constructor(String, byte[])

        /**
         * Returns the attribute name.
         * @return the attribute name
         */
        public String getName() {
            return this.name;
        } // end method 'getName()'

        /**
         * Returns the attribute value.
         * @return the attribute value
         */
        public byte[] getValue() {
            final int len = this.value.length;
            final byte[] res = new byte[len];
            System.arraycopy(this.value, 0, res, 0, len);
            return res;
        } // end method 'getValue()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block c = Block.createBlock(0,
                                              API.createUTF8(this.name),
                                              Value.createFromBlock(Block.createString(this.value)));
            final Block b = Block.createBlock(0,
                                              HASH_UNKNOWN,
                                              Value.createFromBlock(c));
            return Value.createFromBlock(b);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public final boolean equals(final Object obj) {
            if (obj instanceof Unknown) {
                final Unknown that = (Unknown) obj;
                return this.name.equals(that.name)
                    && Arrays.equals(this.value, that.value);
            } else {
                return false;
            } // end if/else
        } // end method 'equals(Object)'

    } // end inner-class 'Unknown'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Attribute fromValue(final Value v) {
        assert v != null : "null v";
        final Block b = v.asBlock();
        final int hash = b.get(0).asLong();
        if (hash == HASH_ANNOTATION_DEFAULT.asLong()) {
            return new AnnotationDefault(Annotation.elementFromValue(b.get(1)));
        } else if (hash == HASH_CODE.asLong()) {
            return new Code(CodeValue.fromValue(b.get(1)));
        } else if (hash == HASH_CONSTANT_VALUE.asLong()) {
            return new ConstantValue(ConstantVal.fromValue(b.get(1)));
        } else if (hash == HASH_DEPRECATED.asLong()) {
            return new Deprecated();
        } else if (hash == HASH_ENCLOSING_METHOD.asLong()) {
            final Block bb = b.get(1).asBlock();
            if (bb.get(1) == Value.ZERO) {
                return new EnclosingMethod(API.stringOfClassName(bb.get(0)));
            } else {
                final Block bbb = bb.get(1).asBlock();
                final Block desc = bbb.get(1).asBlock();
                final List<Descriptor> params = new LinkedList<Descriptor>();
                for (Value p : API.decodeList(desc.get(0))) {
                    params.add(Descriptor.fromValue(v));
                } // end for
                return new EnclosingMethod(API.stringOfClassName(bb.get(0)),
                                           API.stringOfMethodName(bbb.get(0)),
                                           Descriptor.fromValue(desc.get(1)),
                                           params);
            } // end if/else
        } else if (hash == HASH_EXCEPTIONS.asLong()) {
            final List<String> exns = new LinkedList<String>();
            for (Value e : API.decodeList(b.get(1))) {
                exns.add(API.stringOfClassName(e));
            } // end for
            return new Exceptions(exns);
        } else if (hash == HASH_INNER_CLASSES.asLong()) {
            final List<String> inners = new LinkedList<String>();
            final List<String> outers = new LinkedList<String>();
            final List<String> names = new LinkedList<String>();
            final List< List<AccessFlag> > flags = new LinkedList< List<AccessFlag> >();
            for (Value i : API.decodeList(b.get(1))) {
                final Block ii = i.asBlock();
                if (ii.get(0) == Value.ZERO) {
                    inners.add(null);
                } else {
                    inners.add(API.stringOfClassName(ii.get(0).asBlock().get(0)));
                } // end if/else
                if (ii.get(1) == Value.ZERO) {
                    outers.add(null);
                } else {
                    outers.add(API.stringOfClassName(ii.get(1).asBlock().get(0)));
                } // end if/else
                if (ii.get(2) == Value.ZERO) {
                    outers.add(null);
                } else {
                    outers.add(API.stringOfUTF8(ii.get(2).asBlock().get(0)));
                } // end if/else
                final List<AccessFlag> l = new LinkedList<AccessFlag>();
                for (Value f : API.decodeList(ii.get(3))) {
                    l.add(AccessFlag.fromValue(f));
                } // end for
                flags.add(l);
            } // end for
            return new InnerClasses(inners, outers, names, flags);
        } else if (hash == HASH_LINE_NUMBER_TABLE.asLong()) {
            final List<Integer> offsets = new LinkedList<Integer>();
            final List<Integer> lines = new LinkedList<Integer>();
            for (Value c : API.decodeList(b.get(1))) {
                final Block cb = c.asBlock();
                offsets.add(cb.get(0).asLong());
                lines.add(cb.get(1).asLong());
            } // end for
            return new LineNumberTable(offsets, lines);
        } else if (hash == HASH_LOCAL_VARIABLE_TABLE.asLong()) {
            final List<Integer> starts = new LinkedList<Integer>();
            final List<Integer> ends = new LinkedList<Integer>();
            final List<String> names = new LinkedList<String>();
            final List<Descriptor> descriptors = new LinkedList<Descriptor>();
            final List<Integer> indexes = new LinkedList<Integer>();
            for (Value t : API.decodeList(b.get(1))) {
                final Block bb = t.asBlock();
                starts.add(bb.get(0).asLong());
                ends.add(bb.get(1).asLong());
                names.add(API.stringOfUTF8(bb.get(2)));
                descriptors.add(Descriptor.fromValue(bb.get(3)));
                indexes.add(bb.get(4).asLong());
            } // end for
            return new LocalVariableTable(starts, ends, names, descriptors, indexes);
        } else if (hash == HASH_LOCAL_VARIABLE_TYPE_TABLE.asLong()) {
            final List<Integer> starts = new LinkedList<Integer>();
            final List<Integer> ends = new LinkedList<Integer>();
            final List<String> names = new LinkedList<String>();
            final List<String> signatures = new LinkedList<String>();
            final List<Integer> indexes = new LinkedList<Integer>();
            for (Value t : API.decodeList(b.get(1))) {
                final Block bb = t.asBlock();
                starts.add(bb.get(0).asLong());
                ends.add(bb.get(1).asLong());
                names.add(API.stringOfUTF8(bb.get(2)));
                signatures.add(API.stringOfFieldSignature(bb.get(3)));
                indexes.add(bb.get(4).asLong());
            } // end for
            return new LocalVariableTypeTable(starts, ends, names, signatures, indexes);
        } else if (hash == HASH_RUNTIME_INVISIBLE_ANNOTATIONS.asLong()) {
            final List<Annotation> annots = new LinkedList<Annotation>();
            for (Value a : API.decodeList(b.get(1))) {
                annots.add(Annotation.fromValue(a));
            } // end for
            return new RuntimeInvisibleAnnotations(annots);
        } else if (hash == HASH_RUNTIME_INVISIBLE_PARAMETER_ANNOTATIONS.asLong()) {
            final List< List<Annotation> > annotLists = new LinkedList< List<Annotation> >();
            for (Value al : API.decodeList(b.get(1))) {
                final List<Annotation> l = new LinkedList<Annotation>();
                for (Value a : API.decodeList(al)) {
                    l.add(Annotation.fromValue(a));
                } // end for
                annotLists.add(l);
            } // end for
            return new RuntimeInvisibleParameterAnnotations(annotLists);
        } else if (hash == HASH_RUNTIME_VISIBLE_ANNOTATIONS.asLong()) {
            final List<Annotation> annots = new LinkedList<Annotation>();
            for (Value a : API.decodeList(b.get(1))) {
                annots.add(Annotation.fromValue(a));
            } // end for
            return new RuntimeVisibleAnnotations(annots);
        } else if (hash == HASH_RUNTIME_VISIBLE_PARAMETER_ANNOTATIONS.asLong()) {
            final List< List<Annotation> > annotLists = new LinkedList< List<Annotation> >();
            for (Value al : API.decodeList(b.get(1))) {
                final List<Annotation> l = new LinkedList<Annotation>();
                for (Value a : API.decodeList(al)) {
                    l.add(Annotation.fromValue(a));
                } // end for
                annotLists.add(l);
            } // end for
            return new RuntimeVisibleParameterAnnotations(annotLists);
        } else if (hash == HASH_SIGNATURE.asLong()) {
            final Block s = b.get(1).asBlock();
            final int h = s.get(0).asLong();
            if (h == Signature.HASH_CLASS.asLong()) {
                return new Signature(Signature.Kind.CLASS,
                                     API.stringOfClassSignature(s.get(1)));
            } else if (h == Signature.HASH_METHOD.asLong()) {
                return new Signature(Signature.Kind.METHOD,
                                     API.stringOfMethodSignature(s.get(1)));
            } else if (h == Signature.HASH_FIELD.asLong()) {
                return new Signature(Signature.Kind.FIELD,
                                     API.stringOfFieldSignature(s.get(1)));
            } else {
                assert false : "invalid tag";
                return null;
            } // end if/elsif/else
        } else if (hash == HASH_SOURCE_DEBUG_EXTENSION.asLong()) {
            return new SourceDebugExtension(API.stringOfUTF8(b.get(1)));
        } else if (hash == HASH_SOURCE_FILE.asLong()) {
            return new SourceFile(API.stringOfUTF8(b.get(1)));
        } else if (hash == HASH_SYNTHETIC.asLong()) {
            return new Synthetic();
        } else if (hash == HASH_STACK_MAP_TABLE.asLong()) {
            final List<StackMapFrame> frames = new LinkedList<StackMapFrame>();
            for (Value f : API.decodeList(b.get(1))) {
                frames.add(StackMapFrame.fromValue(f));
            } // end for
            return new StackMapTable(frames);
        } else if (hash == HASH_UNKNOWN.asLong()) {
            return new Unknown(API.stringOfUTF8(b.get(1).asBlock().get(0)),
                               b.get(1).asBlock().get(1).asBlock().getBytes());
        } else {
            assert false : "invalid tag";
            return null;
        } // end if/elsif/else
    } // end method 'fromValue(Value)'

} // end class 'Attribute'
