/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Hash;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a version of the Java class file format.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.3
 */
public enum Version implements ToValue {

    /** JDK 1.0 (circa 1996). */
    Java_1_0,

    /** JDK 1.1 (circa 1997). */
    Java_1_1,

    /** J2SE 1.2 (circa 1998). */
    Java_1_2,

    /** J2SE 1.3 (circa 2000). */
    Java_1_3,

    /** J2SE 1.4 (circa 2002). */
    Java_1_4,

    /** J2SE 5.0 (circa 2004). */
    Java_1_5,

    /** Java SE 6 (circa 2006). */
    Java_1_6,

    /** Java SE 7 (to be released). */
    Java_1_7;
    
    /**
     * {@inheritDoc}
     */
    public Value toValue() {
        switch (this) {
        case Java_1_0: return Value.ZERO;
        case Java_1_1: return Value.ONE;
        case Java_1_2: return Value.TWO;
        case Java_1_3: return Value.createFromLong(3);
        case Java_1_4: return Value.createFromLong(4);
        case Java_1_5: return Value.createFromLong(5);
        case Java_1_6: return Value.createFromLong(6);
        case Java_1_7: return Value.createFromLong(7);
        default:
            assert false : "invalid version";
            return Value.ZERO;
        } // end switch
    } // end method 'toValue()'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Version fromValue(final Value v) {
        assert v != null : "null v";
        switch (v.asLong()) {
        case 0: return Java_1_0;
        case 1: return Java_1_1;
        case 2: return Java_1_2;
        case 3: return Java_1_3;
        case 4: return Java_1_4;
        case 5: return Java_1_5;
        case 6: return Java_1_6;
        case 7: return Java_1_7;
        default:
            assert false : "invalid version";
            return null;
        } // end switch
    } // end method 'fromValue(Value)'

    /**
     * Constructs an instance from a string.
     * @param s string to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance, <tt>null</tt> if the string does
     *         not represent a valid version
     */
    static Version fromString(final String s) {
        assert s != null : "null s";
        if (s.equals("1.0")) {
            return Java_1_0;
        } else if (s.equals("1.1")) {
            return Java_1_1;
        } else if (s.equals("1.2")) {
            return Java_1_2;
        } else if (s.equals("1.3")) {
            return Java_1_3;
        } else if (s.equals("1.4")) {
            return Java_1_4;
        } else if (s.equals("1.5")) {
            return Java_1_5;
        } else if (s.equals("1.6")) {
            return Java_1_6;
        } else if (s.equals("1.7")) {
            return Java_1_7;
        } else {
            return null;
        } // end if/elsif/else
    } // end method 'fromString(String)'

} // end class 'Version'
