(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils


(* Types *)

type element =
  | Class of u2
  | Fieldref of u2 * u2
  | Methodref of u2 * u2
  | InterfaceMethodref of u2 * u2
  | String of u2
  | Integer of int32
  | Float of int32
  | Long of int32 * int32
  | Double of int32 * int32
  | NameAndType of u2 * u2
  | UTF8 of utf8

let dummy_element =
  UTF8 (utf8_of_string "Dummy-Constant-Pool-Entry")

type tag =
  | CONSTANT_Class
  | CONSTANT_Fieldref
  | CONSTANT_Methodref
  | CONSTANT_InterfaceMethodref
  | CONSTANT_String
  | CONSTANT_Integer
  | CONSTANT_Float
  | CONSTANT_Long
  | CONSTANT_Double
  | CONSTANT_NameAndType
  | CONSTANT_Utf8

type t = element array


(* Exception *)

type error =
  | Invalid_tag
  | Too_large
  | Invalid_reference
  | Reference_out_of_bounds
  | Dummy_access
  | Malformed_Class_entry
  | Malformed_Fieldref_entry
  | Malformed_Methodref_entry
  | Malformed_InterfaceMethodRef_entry
  | Malformed_String_entry
  | Malformed_NameAndType_entry
  | Not_found

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_tag -> "invalid constant pool tag"
  | Too_large -> "constant pool is too large"
  | Invalid_reference -> "invalid constant pool reference (0)"
  | Reference_out_of_bounds -> "invalid constant pool reference (out of bounds)"
  | Dummy_access -> "access to dummy element"
  | Malformed_Class_entry -> "malformed Class entry"
  | Malformed_Fieldref_entry -> "malformed Fieldref entry"
  | Malformed_Methodref_entry -> "malformed Methodref entry"
  | Malformed_InterfaceMethodRef_entry -> "malformed InterfaceMethodref entry"
  | Malformed_String_entry -> "malformed String entry"
  | Malformed_NameAndType_entry -> "malformed NameAndType entry"
  | Not_found -> "not found"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* I/O functions *)

let element_size = function
  | Class _ -> 1
  | Fieldref _ -> 1
  | Methodref _ -> 1
  | InterfaceMethodref _ -> 1
  | String _ -> 1
  | Integer _ -> 1
  | Float _ -> 1
  | Long _ -> 2
  | Double _ -> 2
  | NameAndType _ -> 1
  | UTF8 _ -> 1

let int_of_tag = function
  | CONSTANT_Class -> 7
  | CONSTANT_Fieldref -> 9
  | CONSTANT_Methodref -> 10
  | CONSTANT_InterfaceMethodref -> 11
  | CONSTANT_String -> 8
  | CONSTANT_Integer -> 3
  | CONSTANT_Float -> 4
  | CONSTANT_Long -> 5
  | CONSTANT_Double -> 6
  | CONSTANT_NameAndType -> 12
  | CONSTANT_Utf8 -> 1

let tag_of_int = function
  | 1 -> CONSTANT_Utf8
  | 3 -> CONSTANT_Integer
  | 4 -> CONSTANT_Float
  | 5 -> CONSTANT_Long
  | 6 -> CONSTANT_Double
  | 7 -> CONSTANT_Class
  | 8 -> CONSTANT_String
  | 9 -> CONSTANT_Fieldref
  | 10 -> CONSTANT_Methodref
  | 11 -> CONSTANT_InterfaceMethodref
  | 12 -> CONSTANT_NameAndType
  | _ -> fail Invalid_tag

let read_element st =
  match tag_of_int (InputStream.read_u1 st) with
  | CONSTANT_Utf8 ->
      let length = InputStream.read_u2 st in
      let bytes = InputStream.read_bytes st length in
      UTF8 (utf8_of_modified (modified_of_bytes bytes))
  | CONSTANT_Integer ->
      let bytes = InputStream.read_s4 st in
      Integer bytes
  | CONSTANT_Float ->
      let bytes = InputStream.read_s4 st in
      Float bytes
  | CONSTANT_Long ->
      let high_bytes = InputStream.read_s4 st in
      let low_bytes = InputStream.read_s4 st in
      Long (high_bytes, low_bytes)
  | CONSTANT_Double ->
      let high_bytes = InputStream.read_s4 st in
      let low_bytes = InputStream.read_s4 st in
      Double (high_bytes, low_bytes)
  | CONSTANT_Class ->
      let name_index = (InputStream.read_u2 st) in
      Class name_index
  | CONSTANT_String ->
      let string_index = (InputStream.read_u2 st) in
      String string_index
  | CONSTANT_Fieldref ->
      let class_index = InputStream.read_u2 st in
      let name_and_type_index = InputStream.read_u2 st in
      Fieldref (class_index, name_and_type_index)
  | CONSTANT_Methodref ->
      let class_index = InputStream.read_u2 st in
      let name_and_type_index = InputStream.read_u2 st in
      Methodref (class_index, name_and_type_index)
  | CONSTANT_InterfaceMethodref ->
      let class_index = InputStream.read_u2 st in
      let name_and_type_index = InputStream.read_u2 st in
      InterfaceMethodref (class_index, name_and_type_index)
  | CONSTANT_NameAndType ->
      let name_index = InputStream.read_u2 st in
      let descriptor_index = InputStream.read_u2 st in
      NameAndType (name_index, descriptor_index)

let read st sz =
  if (sz >= 0) && (sz < 65536) then
    let res = Array.make sz dummy_element in
    let i = ref 1 in
    while !i < sz do
      let e = read_element st in
      res.(!i) <- e;
      i := !i + (element_size e)
    done;
    res
  else
    fail Too_large

let write_element st e =
  if e != dummy_element then
    match e with
    | Class x ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Class);
        OutputStream.write_u2 st x
    | Fieldref (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Fieldref);
        OutputStream.write_u2 st x;
        OutputStream.write_u2 st y
    | Methodref (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Methodref);
        OutputStream.write_u2 st x;
        OutputStream.write_u2 st y
    | InterfaceMethodref (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_InterfaceMethodref);
        OutputStream.write_u2 st x;
        OutputStream.write_u2 st y
    | String x ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_String);
        OutputStream.write_u2 st x
    | Integer x ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Integer);
        OutputStream.write_s4 st x
    | Float x ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Float);
        OutputStream.write_s4 st x
    | Long (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Long);
        OutputStream.write_s4 st x;
        OutputStream.write_s4 st y
    | Double (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_Double);
        OutputStream.write_s4 st x;
        OutputStream.write_s4 st y
    | NameAndType (x, y) ->
        OutputStream.write_u1 st (int_of_tag CONSTANT_NameAndType);
        OutputStream.write_u2 st x;
        OutputStream.write_u2 st y
    | UTF8 x ->
        let bytes = bytes_of_modified (modified_of_utf8 x) in
        OutputStream.write_u1 st (int_of_tag CONSTANT_Utf8);
        OutputStream.write_u2 st (String.length bytes);
        OutputStream.write_bytes st bytes

let write st pool =
  if Array.length pool < 65536 then
    Array.iter (write_element st) pool
  else
    fail Too_large


(* Check functions *)

let get_entry pool i =
  if i = 0 then
    fail Invalid_reference
  else if (i < 0) || (i >= (Array.length pool)) then
    fail Reference_out_of_bounds
  else
    let res = pool.(i) in
    if res == dummy_element then
      fail Dummy_access
    else
      res

let check pool =
  let get_entry = get_entry pool in
  let check_entry = function
    | Class name_index ->
        (match get_entry name_index with
        | UTF8 _ -> ()
        | _ -> fail Malformed_Class_entry)
    | Fieldref (class_index, name_and_type_index) ->
        (match (get_entry class_index), (get_entry name_and_type_index) with
        | (Class _), (NameAndType _) -> ()
        | _ -> fail Malformed_Fieldref_entry)
    | Methodref (class_index, name_and_type_index) ->
        (match (get_entry class_index), (get_entry name_and_type_index) with
        | (Class _), (NameAndType _) -> ()
        | _ -> fail Malformed_Methodref_entry)
    | InterfaceMethodref (class_index, name_and_type_index) ->
        (match (get_entry class_index), (get_entry name_and_type_index) with
        | (Class _), (NameAndType _) -> ()
        | _ -> fail Malformed_InterfaceMethodRef_entry)
    | String value_index ->
        (match get_entry value_index with
        | UTF8 _ -> ()
        | _ -> fail Malformed_String_entry)
    | Integer _ -> ()
    | Float _ -> ()
    | Long _ -> ()
    | Double _ -> ()
    | NameAndType (name_index, desc_index) ->
        (match (get_entry name_index), (get_entry desc_index) with
        | (UTF8 _), (UTF8 _) -> ()
        | _ -> fail Malformed_NameAndType_entry)
    | UTF8 _ -> () in
  Array.iter check_entry pool


(* Extendable pools *)

type extendable = {
    mutable current : element array;
    mutable next : int;
  }

let make_extendable () =
  { current = Array.make 128 dummy_element;
    next = 1; }

let make_extendable_from_pool pool =
  let len = Array.length pool in
  let size = ref 128 in
  while (!size < 65536) && (!size < len) do
    size := 2 * !size
  done;
  let size' = min 65535 !size in
  if (size' >= len) then
    { current = Array.init size' (fun i -> if i < len then pool.(i) else dummy_element) ;
      next = len; }
  else
    fail Too_large

let get_extendable_entry pool i =
  if i = 0 then
    fail Invalid_reference
  else if (i < 0) || (i >= pool.next) then
    fail Reference_out_of_bounds
  else
    let res = pool.current.(i) in
    if res == dummy_element then
      fail Dummy_access
    else
      res

let rec add_element ext e =
  let len = (Array.length ext.current) in
  let size = element_size e in
  let next = ext.next in
  if next + size <= len then begin
    ext.current.(ext.next) <- e;
    ext.next <- next + size;
    next
  end else begin
    if len >= 65535 then fail Too_large;
    let new_array = Array.init
        (min 65535 (2 * len))
        (fun i -> if i < len then ext.current.(i) else dummy_element) in
    ext.current <- new_array;
    add_element ext e
  end

let index_of_element ext elem =
  let index_with p a =
    let len = Array.length a in
    let idx = ref 0 in
    while (!idx < len) && not (p a.(!idx)) do
      incr idx
    done;
    if (!idx < len) then
      !idx
    else
      fail Not_found in
  index_with
    (fun x ->
      (x != dummy_element)
        && (match x, elem with
        | (UTF8 u), (UTF8 u') -> utf8_eq u u'
        | _ -> x = elem))
    ext.current

let add_if_not_found ext elem =
  try
    index_of_element ext elem
  with Exception _ ->
    add_element ext elem

let add_utf8 ext s =
  let elem = (UTF8 s) in
  add_if_not_found ext elem

let add_class ext n =
  let name_index = add_utf8 ext (Name.internal_utf8_for_class n) in
  let elem = Class name_index in
  add_if_not_found ext elem

let add_array_class ext d =
  let name_index = add_utf8 ext (Descriptor.utf8_of_java_type (d :> Descriptor.java_type)) in
  let elem = Class name_index in
  add_if_not_found ext elem

let add_name_and_type ext n t =
  let name_index = add_utf8 ext n in
  let type_index = add_utf8 ext t in
  let elem = NameAndType (name_index, type_index) in
  add_if_not_found ext elem

let add_field ext cn n t =
  let class_index = add_class ext cn in
  let d = Descriptor.utf8_of_field t in
  let name_and_type_index = add_name_and_type ext (Name.utf8_for_field n) d in
  let elem = Fieldref (class_index, name_and_type_index) in
  add_if_not_found ext elem

let add_method ext cn n t =
  let class_index = add_class ext cn in
  let d = Descriptor.utf8_of_method t in
  let name_and_type_index = add_name_and_type ext (Name.utf8_for_method n) d in
  let elem = Methodref (class_index, name_and_type_index) in
  add_if_not_found ext elem

let add_interface_method ext cn n t =
  let class_index = add_class ext cn in
  let d = Descriptor.utf8_of_method t in
  let name_and_type_index = add_name_and_type ext (Name.utf8_for_method n) d in
  let elem = InterfaceMethodref (class_index, name_and_type_index) in
  add_if_not_found ext elem

let add_array_method ext at n t =
  let class_index = add_array_class ext at in
  let d = Descriptor.utf8_of_method t in
  let name_and_type_index = add_name_and_type ext (Name.utf8_for_method n) d in
  let elem = Methodref (class_index, name_and_type_index) in
  add_if_not_found ext elem

let add_string ext s =
  let v = add_utf8 ext s in
  let elem = String v in
  add_if_not_found ext elem

let add_integer ext i =
  let elem = Integer i in
  add_if_not_found ext elem

let add_float ext f =
  let elem = Float (Int32.bits_of_float f) in
  add_if_not_found ext elem

let add_long ext l =
  let high = Int64.to_int32 (Int64.shift_right_logical l 32) in
  let low = Int64.to_int32 (Int64.logand l 0x00000000FFFFFFFFL) in
  let elem = Long (high, low) in
  add_if_not_found ext elem

let add_double ext d =
  let l = Int64.bits_of_float d in
  let high = Int64.to_int32 (Int64.shift_right_logical l 32) in
  let low = Int64.to_int32 (Int64.logand l 0x00000000FFFFFFFFL) in
  let elem = Double (high, low) in
  add_if_not_found ext elem

let to_pool ext =
  Array.sub ext.current 0 ext.next
