(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Java types definition *)

type java_type =
  [ `Boolean
  | `Byte
  | `Char
  | `Double
  | `Float
  | `Int
  | `Long
  | `Short
  | `Void
  | `Class of Name.for_class
  | `Array of 'a ] constraint 'a = [ `Boolean
                                   | `Byte
                                   | `Char
                                   | `Double
                                   | `Float
                                   | `Int
                                   | `Long
                                   | `Short
                                   | `Class of Name.for_class
                                   | `Array of 'a ]

type array_type =
  [ `Array of 'a ] constraint 'a = [ `Boolean
                                   | `Byte
                                   | `Char
                                   | `Double
                                   | `Float
                                   | `Int
                                   | `Long
                                   | `Short
                                   | `Class of Name.for_class
                                   | `Array of 'a ]


(* Exception *)

type error =
  | Invalid_class_name
  | Invalid_array_element_type
  | Array_with_too_many_dimensions
  | Invalid_descriptor_string
  | Empty_descriptor_string
  | Invalid_field_type
  | Invalid_local_variable_type
  | Invalid_method_descriptor
  | Invalid_method_parameter_type
  | Void_not_allowed

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_class_name -> "invalid class name"
  | Invalid_array_element_type -> "invalid array element type (void)"
  | Array_with_too_many_dimensions -> "array with more than 255 dimensions"
  | Invalid_descriptor_string -> "invalid descriptor string"
  | Empty_descriptor_string -> "empty descriptor string"
  | Invalid_field_type -> "invalid field type (void)"
  | Invalid_local_variable_type -> "invalid local variable type (void)"
  | Invalid_method_descriptor -> "invalid method descriptor"
  | Invalid_method_parameter_type -> "invalid parameter type (void)"
  | Void_not_allowed -> "void is not allowed here"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Utility functions *)

let is_primitive = function
  | `Boolean
  | `Byte
  | `Char
  | `Double
  | `Float
  | `Int
  | `Long
  | `Short -> true
  | `Void
  | `Class _
  | `Array _ -> false

let filter_void err = function
  | `Boolean -> `Boolean
  | `Byte -> `Byte
  | `Char -> `Char
  | `Double -> `Double
  | `Float -> `Float
  | `Int -> `Int
  | `Long -> `Long
  | `Short -> `Short
  | `Void -> fail err
  | `Class c -> `Class c
  | `Array t -> `Array t

let filter_array err = function
  | `Boolean
  | `Byte
  | `Char
  | `Double
  | `Float
  | `Int
  | `Long
  | `Short
  | `Void
  | `Class _ -> fail err
  | `Array t -> `Array t

let java_type_of_partial_utf8 str i =
  let len = utf8_length str in
  let rec jtopu n idx =
    if idx < len then begin
      let ch = utf8_get str idx in
      if uchar_eq ch capital_z then
        (`Boolean, succ idx)
      else if uchar_eq ch capital_b then
        (`Byte, succ idx)
      else if uchar_eq ch capital_c then
        (`Char, succ idx)
      else if uchar_eq ch capital_d then
        (`Double, succ idx)
      else if uchar_eq ch capital_f then
        (`Float, succ idx)
      else if uchar_eq ch capital_i then
        (`Int, succ idx)
      else if uchar_eq ch capital_j then
        (`Long, succ idx)
      else if uchar_eq ch capital_s then
        (`Short, succ idx)
      else if uchar_eq ch capital_v then
        (`Void, succ idx)
      else if uchar_eq ch capital_l then
        try
          let index = utf8_index_from str (succ idx) semi_colon in
          let name = (Name.make_for_class_from_internal (utf8_substring str (succ idx) (pred index))) in
          (`Class name, index + 1)
        with
        | Not_found
        | Name.Exception _ -> fail Invalid_class_name
      else if uchar_eq ch opening_square_bracket then
        if n < 255 then
          let t, res = jtopu (succ n) (succ idx) in
          (`Array (filter_void Invalid_array_element_type t), res)
        else
          fail Array_with_too_many_dimensions
      else
        fail Invalid_descriptor_string
    end else
      fail Empty_descriptor_string in
  jtopu 0 i

let java_type_of_utf8 s =
  let res, idx = java_type_of_partial_utf8 s 0 in
  if idx = utf8_length s then
    res
  else
    fail Invalid_descriptor_string

let utf8_of_java_type =
  let rec uojt n = function
    | `Boolean -> utf8_of_string "Z"
    | `Byte -> utf8_of_string "B"
    | `Char -> utf8_of_string "C"
    | `Double -> utf8_of_string "D"
    | `Float -> utf8_of_string "F"
    | `Int -> utf8_of_string "I"
    | `Long -> utf8_of_string "J"
    | `Short -> utf8_of_string "S"
    | `Void -> utf8_of_string "V"
    | `Class c ->
        (utf8_of_string "L")
          ++ (Name.internal_utf8_for_class c)
          ++ (utf8_of_string ";")
    | `Array jt ->
        if n < 255 then
          (utf8_of_string "[") ++ (uojt (succ n) (jt :> java_type))
        else
          fail Array_with_too_many_dimensions
  in uojt 0


(* Field descriptors *)

type for_field =
  [ `Boolean
  | `Byte
  | `Char
  | `Double
  | `Float
  | `Int
  | `Long
  | `Short
  | `Class of Name.for_class
  | `Array of 'a ] constraint 'a = [ `Boolean
                                   | `Byte
                                   | `Char
                                   | `Double
                                   | `Float
                                   | `Int
                                   | `Long
                                   | `Short
                                   | `Class of Name.for_class
                                   | `Array of 'a ]

let field_of_utf8 str =
  let t = java_type_of_utf8 str in
  filter_void Invalid_field_type t

let utf8_of_field fd =
  utf8_of_java_type (fd :> java_type)


(* Method descriptors *)

type for_parameter = for_field

let parameter_of_utf8 = field_of_utf8

let utf8_of_parameter = utf8_of_field

type for_method = (for_parameter list) * java_type

let method_of_utf8 str =
  let len = utf8_length str in
  if (len > 2) && (uchar_eq opening_parenthesis (utf8_get str 0)) then
    let index = (try
      utf8_index_from str 1 closing_parenthesis
    with Not_found -> fail Invalid_method_descriptor) in
    let (ret, last) = java_type_of_partial_utf8 str (index + 1) in
    if (last <> len) then fail Invalid_method_descriptor;
    let params = ref [] in
    let curr = ref 1 in
    while !curr < index do
      let t, i = java_type_of_partial_utf8 str !curr in
      params := (filter_void Invalid_method_parameter_type t) :: !params;
      curr := i
    done;
    if !curr = index then
      ((List.rev !params), ret)
    else
      fail Invalid_method_descriptor
  else
    fail Invalid_method_descriptor

let utf8_of_method (params, return) =
  (utf8_of_string "(")
    ++ (utf8_concat (List.map utf8_of_parameter params))
    ++ (utf8_of_string ")")
    ++ (utf8_of_java_type return)
