(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(* Type *)

type t =
  | To_buffer of Buffer.t * bool ref
  | To_channel of out_channel


(* Exception *)

type error =
  | Unable_to_write_data
  | Unable_to_close_stream

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Unable_to_write_data -> "unable to write data"
  | Unable_to_close_stream -> "unable to close stream"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Constructors *)

let make_of_buffer buf =
  To_buffer (buf, ref true)

let make_of_channel ch =
  set_binary_mode_out ch true;
  To_channel ch


(* Functions *)

let write_u1 st x =
  match st with
  | To_buffer (buf, o) ->
      if !o then
        Buffer.add_char buf (char_of_int (x land 0x000000FF))
      else
        fail Unable_to_write_data
  | To_channel ch ->
      try output_byte ch x with _ -> fail Unable_to_write_data

let write_u2 st x =
  let x1 = (x lsr 8) land 0x000000FF in
  let x2 = x land 0x000000FF in
  write_u1 st x1;
  write_u1 st x2

let write_u4 st x =
  let x1 = Int64.logand (Int64.shift_right_logical x 24) 0x000000FFL in
  let x2 = Int64.logand (Int64.shift_right_logical x 16) 0x000000FFL in
  let x3 = Int64.logand (Int64.shift_right_logical x 8) 0x000000FFL in
  let x4 = Int64.logand x 0x000000FFL in
  write_u1 st (Int64.to_int x1);
  write_u1 st (Int64.to_int x2);
  write_u1 st (Int64.to_int x3);
  write_u1 st (Int64.to_int x4)

let write_s1 st x =
  let x' = x land 0x000000FF in
  write_u1 st x'

let write_s2 = write_u2

let write_s4 st x =
  let x1 = Int32.logand (Int32.shift_right_logical x 24) 0x000000FFl in
  let x2 = Int32.logand (Int32.shift_right_logical x 16) 0x000000FFl in
  let x3 = Int32.logand (Int32.shift_right_logical x 8) 0x000000FFl in
  let x4 = Int32.logand x 0x000000FFl in
  write_u1 st (Int32.to_int x1);
  write_u1 st (Int32.to_int x2);
  write_u1 st (Int32.to_int x3);
  write_u1 st (Int32.to_int x4)

let write_bytes_from st s pos len =
  match st with
  | To_buffer (buf, o) ->
      if !o then
        Buffer.add_substring buf s pos len
      else
        fail Unable_to_write_data
  | To_channel ch ->
      try output ch s pos len with _ -> fail Unable_to_write_data

let write_bytes st s =
  write_bytes_from st s 0 (String.length s)

let close = function
  | To_buffer (_, o) -> o := false
  | To_channel c -> try close_out c with _ -> fail Unable_to_close_stream
