(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Constants *)

let opening_bracket = utf8_of_string "<"

let comma_space = utf8_of_string ", "

let closing_bracket = utf8_of_string ">"

let closing_bracket_space = utf8_of_string "> "

let class' = utf8_of_string "class "

let interface = utf8_of_string "interface "

let extends = utf8_of_string " extends "

let implements = utf8_of_string " implements "

let throws = utf8_of_string " throws "

let space_and_space = utf8_of_string " & "

let array_suffix = utf8_of_string "[]"

let wildcard = utf8_of_string "?"

let wildcard_extends = utf8_of_string "? extends "

let wildcard_super = utf8_of_string "? super "

let true_string = utf8_of_string "true"

let false_string = utf8_of_string "false"

let dot_string = utf8_of_string "."

let equal = utf8_of_string "="

let tab = utf8_of_string "  "

let opening_square_bracket = utf8_of_string "["

let closing_square_bracket = utf8_of_string "]"

let at_string = utf8_of_string "@"

let opening_parenthesis = utf8_of_string "("

let closing_parenthesis = utf8_of_string ")"

let semi_colon = utf8_of_string ";"

let opening_curly_bracket = utf8_of_string " {"

let closing_curly_bracket = utf8_of_string "}"

let static_block = utf8_of_string "static"


(* Functions *)

let printed_flag = function
  | `Public
  | `Private
  | `Protected
  | `Static
  | `Final
  | `Synchronized
  | `Volatile
  | `Abstract
  | `Strict
  | `Transient
  | `Native -> true
  | `Super
  | `Bridge
  | `Varargs
  | `Interface
  | `Synthetic
  | `Annotation
  | `Enum -> false

let rec extract_exceptions = function
  | (`Exceptions e) :: _ -> e
  | _ :: tl -> extract_exceptions tl
  | [] -> raise Not_found

let extract_annotations l =
  let rec extract accu = function
    | (`RuntimeVisibleAnnotations a) :: tl
    | (`RuntimeInvisibleAnnotations a) :: tl ->
        extract (accu @ a) tl
    | _ :: tl -> extract accu tl
    | [] -> accu in
  extract [] l

let rec utf8_of_class_signature cs =
  let ftp =
    if cs.Signature.formal_type_parameters <> [] then
      opening_bracket
        ++ (utf8_concat_sep
              comma_space
              (List.map utf8_of_formal_type_parameter cs.Signature.formal_type_parameters))
        ++ closing_bracket
    else empty_utf8 in
  let scs = utf8_of_class_type_signature cs.Signature.super_class_signature in
  let sis = List.map utf8_of_class_type_signature cs.Signature.super_interface_signatures in
  (ftp, scs, sis)
and utf8_of_formal_type_parameter ftp =
  let fts = utf8_of_field_type_signature ftp.Signature.class_bound in
  ftp.Signature.identifier
    ++ (if utf8_eq fts java_lang_Object then
      (if ftp.Signature.interface_bounds <> [] then
        extends
          ++ (utf8_concat_sep
                space_and_space
                (List.map utf8_of_field_type_signature ftp.Signature.interface_bounds))
      else
        empty_utf8)
    else
      extends
        ++ fts
        ++ (utf8_concat
              (List.map
                 (fun x -> space_and_space ++ (utf8_of_field_type_signature x))
                 ftp.Signature.interface_bounds)))
and utf8_of_field_type_signature = function
  | Signature.Class_type_signature cts -> utf8_of_class_type_signature cts
  | Signature.Array_type_signature ats -> utf8_of_array_type_signature ats
  | Signature.Type_variable_signature tvs -> utf8_of_type_variable_signature tvs
and utf8_of_class_type_signature cts =
  (Name.printable_utf8_for_class cts.Signature.qualified_class_name)
    ++ (if cts.Signature.type_arguments <> [] then
      opening_bracket
        ++ (utf8_concat_sep
              comma_space
              (List.map utf8_of_type_argument cts.Signature.type_arguments))
        ++ closing_bracket
    else empty_utf8)
and utf8_of_array_type_signature ats =
  (utf8_of_type_signature ats) ++ array_suffix
and utf8_of_type_signature = function
  | Signature.Field_type_signature fts -> utf8_of_field_type_signature fts
  | Signature.Base_type jt -> Source.utf8_of_java_type jt
and utf8_of_type_variable_signature tvs = tvs
and utf8_of_type_argument = function
  | Signature.Star -> wildcard
  | Signature.Plus fts ->
      wildcard_extends ++ (utf8_of_field_type_signature fts)
  | Signature.Minus fts ->
      wildcard_super ++ (utf8_of_field_type_signature fts)
  | Signature.Simple fts ->
      utf8_of_field_type_signature fts
and utf8_of_class_type_signature_suffix ctss = utf8_of_class_type_signature ctss
and utf8_of_throws_signature = function
  | Signature.Throws_class_type_signature cts ->
      utf8_of_class_type_signature cts
  | Signature.Throws_type_variable_signature tvs ->
      utf8_of_type_variable_signature tvs

let rec utf8_of_element_value = function
  | Annotation.Boolean_value true -> true_string
  | Annotation.Boolean_value false -> false_string
  | Annotation.Byte_value i -> utf8_of_string (string_of_int i)
  | Annotation.Char_value c -> utf8_escape_char c
  | Annotation.Double_value f -> utf8_of_string (string_of_float f)
  | Annotation.Float_value f -> utf8_of_string (string_of_float f)
  | Annotation.Int_value i -> utf8_of_string (Int32.to_string i)
  | Annotation.Long_value i -> utf8_of_string (Int64.to_string i)
  | Annotation.Short_value i -> utf8_of_string (string_of_int i)
  | Annotation.String_value s -> utf8_escape s
  | Annotation.Enum_value (c, i) ->
      (Name.printable_utf8_for_class c)
        ++ dot_string
        ++ (Name.utf8_for_field i)
  | Annotation.Class_value cn -> Name.printable_utf8_for_class cn
  | Annotation.Annotation_value a -> utf8_of_annotation a
  | Annotation.Array_value l  ->
      opening_square_bracket ++
        (utf8_concat_sep comma_space (List.map utf8_of_element_value l)) ++
        closing_square_bracket
and utf8_of_pair (n, v) =
  n ++ equal ++ (utf8_of_element_value v)
and utf8_of_annotation (name, values) =
  at_string ++ (Name.printable_utf8_for_class name) ++
  (if values <> [] then
    opening_parenthesis ++
      (utf8_concat_sep comma_space (List.map utf8_of_pair values)) ++
      closing_parenthesis
  else
    empty_utf8)

let add_annotations buffer prefix annotations =
  List.iter
    (fun a -> utf8_buf_add_endline buffer (prefix ++ (utf8_of_annotation a)))
    annotations

let add_parents buffer is_interface cd =
  try
    let rec extract_signature = function
      | (`Signature (`Class s)) :: _ -> s
      | _ :: tl -> extract_signature tl
      | [] -> raise Not_found in
    let signature = extract_signature cd.ClassDefinition.attributes in
    let cs, sc, si = utf8_of_class_signature signature in
    utf8_buf_add_string buffer cs;
    if not is_interface then begin
      utf8_buf_add_string buffer extends;
      utf8_buf_add_string buffer sc
    end;
    if si <> [] then begin
      utf8_buf_add_string buffer (if is_interface then extends else implements);
      utf8_buf_add_string buffer (utf8_concat_sep comma_space si)
    end
  with Not_found ->
    (match is_interface, cd.ClassDefinition.extends with
    | false, (Some v) ->
        let name = Name.printable_utf8_for_class v in
        utf8_buf_add_string buffer extends;
        utf8_buf_add_string buffer name;
    | _ -> ());
    if cd.ClassDefinition.implements <> [] then begin
      utf8_buf_add_string buffer (if is_interface then extends else implements);
      utf8_buf_add_string
        buffer
        (utf8_concat_sep
           comma_space
           (List.map
              Name.printable_utf8_for_class
              cd.ClassDefinition.implements))
    end

let add_field buffer (flags, name, desc, attrs) =
  add_annotations buffer tab (extract_annotations attrs);
  utf8_buf_add_string buffer tab;
  utf8_buf_add_string
    buffer
    (Source.utf8_of_flags (List.filter printed_flag (flags :> AccessFlag.t list)));
  (try
    let rec extract_signature = function
      | (`Signature (`Field s)) :: _ -> s
      | _ :: tl -> extract_signature tl
      | [] -> raise Not_found in
    let fts = extract_signature attrs in
    utf8_buf_add_string buffer (utf8_of_field_type_signature fts)
  with Not_found ->
    utf8_buf_add_string buffer (Source.utf8_of_java_type (desc :> Descriptor.java_type)));
  utf8_buf_add_char buffer (uchar_of_char ' ');
  utf8_buf_add_string buffer (Name.utf8_for_field name);
  utf8_buf_add_endline buffer semi_colon

let add_method buffer class_name m =
  let flags, name, desc, attrs =
    match m with
    | Method.Regular (f, n, d, a) ->
        (f, (Name.utf8_for_method n), d, a)
    | Method.Constructor (f, d, a) ->
        ((f :> AccessFlag.for_method list), class_constructor, (d, `Void), a)
    | Method.Initializer (s, a) ->
        ((if s then [`Strict] else []), class_initializer, ([], `Void), a) in
  add_annotations buffer tab (extract_annotations attrs);
  utf8_buf_add_string buffer tab;
  utf8_buf_add_string
    buffer
    (Source.utf8_of_flags (List.filter printed_flag (flags :> AccessFlag.t list)));
  try
    let rec extract_signature = function
      | (`Signature (`Method s)) :: _ -> s
      | _ :: tl -> extract_signature tl
      | [] -> raise Not_found in
    let fts = extract_signature attrs in
    if fts.Signature.formal_type_params <> [] then begin
      utf8_buf_add_char buffer (uchar_of_char '<');
      utf8_buf_add_string
        buffer
        (utf8_concat_sep
           comma_space
           (List.map utf8_of_formal_type_parameter fts.Signature.formal_type_params));
      utf8_buf_add_string buffer closing_bracket_space
    end;
    if utf8_eq name class_initializer then
      utf8_buf_add_string buffer static_block
    else if utf8_eq name class_constructor then
      utf8_buf_add_string buffer class_name
    else begin
      utf8_buf_add_string buffer (utf8_of_type_signature fts.Signature.return);
      utf8_buf_add_char buffer (uchar_of_char ' ');
      utf8_buf_add_string buffer name
    end;
    utf8_buf_add_char buffer (uchar_of_char '(');
    utf8_buf_add_string
      buffer
      (utf8_concat_sep
         comma_space
         (List.map utf8_of_type_signature fts.Signature.types));
    utf8_buf_add_char buffer (uchar_of_char ')');
    if fts.Signature.throws_signatures <> [] then begin
      utf8_buf_add_string buffer throws;
      utf8_buf_add_string
        buffer
        (utf8_concat_sep
           comma_space
           (List.map utf8_of_throws_signature fts.Signature.throws_signatures))
    end;
    utf8_buf_add_endline buffer semi_colon
  with Not_found ->
    let params, return = desc in
    if utf8_eq name class_initializer then
      utf8_buf_add_string buffer static_block
    else if utf8_eq name class_constructor then
      utf8_buf_add_string buffer class_name
    else begin
      utf8_buf_add_string buffer (Source.utf8_of_java_type return);
      utf8_buf_add_char buffer (uchar_of_char ' ');
      utf8_buf_add_string buffer name
    end;
    utf8_buf_add_char buffer (uchar_of_char '(');
    utf8_buf_add_string
      buffer
      (utf8_concat_sep
         comma_space
         (List.map Source.utf8_of_java_type (params :> Descriptor.java_type list)));
    utf8_buf_add_char buffer (uchar_of_char ')');
    (try
      let thrown = extract_exceptions attrs in
      utf8_buf_add_string buffer throws;
      utf8_buf_add_string
        buffer
        (utf8_concat_sep
           comma_space
           (List.map (fun x -> Name.printable_utf8_for_class x) thrown))
    with Not_found -> ());
    utf8_buf_add_endline buffer semi_colon

let print_to_buffer buffer cp s =
  let cl = ClassLoader.make cp in
  let cd = ClassLoader.find cl s in
  add_annotations buffer empty_utf8 (extract_annotations cd.ClassDefinition.attributes);
  let flags = (cd.ClassDefinition.access_flags :> AccessFlag.t list) in
  let is_interface = List.mem `Interface flags in
  let flags_to_print = List.filter (fun f ->
    if f = `Abstract then
      not is_interface
    else
      printed_flag f) flags in
  utf8_buf_add_string
    buffer
    (Source.utf8_of_flags (flags_to_print :> AccessFlag.t list));
  let name = Name.printable_utf8_for_class cd.ClassDefinition.name in
  utf8_buf_add_string
    buffer
    (if is_interface then interface else class');
  utf8_buf_add_string buffer name;
  add_parents buffer is_interface cd;
  utf8_buf_add_endline buffer opening_curly_bracket;
  List.iter
    (add_field buffer)
    (List.sort Source.field_order cd.ClassDefinition.fields);
  List.iter
    (add_method buffer name)
    (List.sort Source.method_order cd.ClassDefinition.methods);
  utf8_buf_add_string buffer closing_curly_bracket

let print_to_channel chan cp s =
  let buffer = utf8_buf_make () in
  print_to_buffer buffer cp s;
  Source.output chan buffer

let print cp s =
  print_to_channel stdout cp s
