(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module provides some utility functions related to source handling. *)


(** {6 Exception} *)

type error =
  | Invalid_array_dimensions
  | Invalid_type
  | Unknown_flag

exception Exception of error
(** Exception to be raised when a function of this module fails.  *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 Orders over some Java elements} *)

val flag_order : AccessFlag.t -> AccessFlag.t -> int
(** Order over flags. *)

val field_order : Field.t -> Field.t -> int
(** Order over fields. *)

val meth_desc_order : Descriptor.for_method -> Descriptor.for_method -> int
(** Order over method descriptors. *)

val method_order : Method.t -> Method.t -> int
(** Order over methods. *)

val attribute_order : Attribute.t -> Attribute.t -> int
(** Order over attributes. *)


(** {6 Some conversion functions} *)

val utf8_of_java_type : Descriptor.java_type -> Utils.utf8
(** Returns the textual representation of the passed Java type. *)

val java_type_of_utf8 : Utils.utf8 -> Descriptor.java_type
(** Returns the Java type corresponding to the passed string.
   Raises [Exception] if the string does not represent a Java type.
   Also Raises [Descriptor.Exception] if the type is invalid. *)

val java_type_of_utf8_no_void : Utils.utf8 -> Descriptor.for_field
(** Same as [java_type_of_utf8] but raises [Descriptor.Exception] if the
   decoded type is equal to the Java type {i void}. *)

val utf8_of_flag : AccessFlag.t -> Utils.utf8
(** Converts the passed flag to its corresponding string. *)

val utf8_of_flags : AccessFlag.t list -> Utils.utf8
(** Converts the passed flag list to its corresponding string.
   Flags are separated by a single space, and a single space is also added at
   the end of the returned string. *)

val flag_of_utf8 : Utils.utf8 -> AccessFlag.t
(** Converts the passed string to its corresponding flag.
   Raises [Exception] if passed string is not a valid flag string. *)


(** {6 Miscellaneous} *)

val output : out_channel -> Utils.utf8_buf -> unit
(** Outputs the content of the passed buffer onto the passed channel.
   An exception is raised if an error occurs during the write operation. *)
