(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open CamomileLibrary


(* Integer types *)

type u1 = int

type s1 = int

type u2 = int

type s2 = int

type u4 = int64

type s4 = int32


(* Exceptions *)

type unicode_error =
  | Unrepresentable_character
  | Invalid_character_code
  | Unable_to_convert_to_modified_utf8
  | Unable_to_convert_from_modified_utf8
  | Unable_to_convert_to_utf8
  | Unable_to_convert_from_utf8
  | Invalid_index
  | Index_out_of_bounds
  | Invalid_escaped_string

exception Unicode_exception of unicode_error

let fail_unicode e = raise (Unicode_exception e)

let string_of_unicode_error = function
  | Unrepresentable_character -> "unrepresentable character"
  | Invalid_character_code -> "invalid character code"
  | Unable_to_convert_to_modified_utf8 -> "unable to convert to modified UTF8"
  | Unable_to_convert_from_modified_utf8 -> "unable to convert from modified UTF8"
  | Unable_to_convert_to_utf8 -> "unable to convert to UTF8"
  | Unable_to_convert_from_utf8 -> "unable to convert from UTF8"
  | Invalid_index -> "invalid index"
  | Index_out_of_bounds -> "index out of bounds"
  | Invalid_escaped_string -> "invalid escaped string"

let () =
  Printexc.register_printer
    (function
      | Unicode_exception e -> Some (string_of_unicode_error e)
      | _ -> None)

type lexer_error =
  | End_of_lexer
  | Invalid_consume of char * char

exception Lexer_exception of lexer_error

let fail_lexer e = raise (Lexer_exception e)

let string_of_lexer_error = function
  | End_of_lexer -> "end of lexer"
  | Invalid_consume (w, f) -> Printf.sprintf "invalid consume ('%c' waited but '%c' found)" w f

let () =
  Printexc.register_printer
    (function
      | Lexer_exception e -> Some (string_of_lexer_error e)
      | _ -> None)


(* Unicode characters *)

type uchar = UChar.t

let uchar_of_char = UChar.of_char

let char_of_uchar x =
  try
    UChar.char_of x
  with UChar.Out_of_range -> fail_unicode Unrepresentable_character

let uchar_of_code x =
  try
    UChar.chr x
  with Invalid_argument _ -> fail_unicode Invalid_character_code

let code_of_uchar x =
  try
    UChar.code x
  with UChar.Out_of_range -> fail_unicode Unrepresentable_character

let uchar_eq = UChar.eq

module CharInfo = UCharInfo.Make (CamomileDefaultConfig)

let uchar_is_letter ch =
  match CharInfo.general_category ch with
  | `Lu | `Ll | `Lt | `Lo | `Lm -> true
  | _ -> false

let uchar_is_digit ch =
  match CharInfo.general_category ch with
  | `Nd | `Nl | `No -> true
  | _ -> false

let underscore = uchar_of_char '_'

let uchar_is_letter_or_digit ch =
  uchar_eq underscore ch
  || uchar_is_letter ch
  || uchar_is_digit ch


(* UTF8 strings *)

type utf8 = UTF8.t

type modified = string

let make_utf8 l =
  let res = UTF8.Buf.create (List.length l) in
  List.iter (fun x -> UTF8.Buf.add_char res x) l;
  UTF8.Buf.contents res

let modified_of_bytes x = x

let bytes_of_modified x = x

let modified_of_utf8 u =
  try
    UTF8.validate u;
    let b = Buffer.create 16 in
    UTF8.iter
      (fun ch ->
        let code = UChar.uint_code ch in
        if (code >= 0x0001) && (code <= 0x007F) then
          Buffer.add_char b (char_of_int code)
        else if (code = 0x0000) || ((code >= 0x0080) && (code <= 0x07FF)) then begin
          Buffer.add_char b (char_of_int ((0x00C0 lor (code lsr 6))));
          Buffer.add_char b (char_of_int ((0x0080 lor (code land 0x003F))))
        end else if (code >= 0x0800) && (code <= 0xFFFF) then begin
          Buffer.add_char b (char_of_int ((0x00E0 lor (code lsr 12))));
          Buffer.add_char b (char_of_int ((0x0080 lor ((code lsr 6) land 0x003F))));
          Buffer.add_char b (char_of_int ((0x0080 lor (code land 0x003F))))
        end else begin
          Buffer.add_char b (char_of_int 0x00ED);
          Buffer.add_char b (char_of_int ((0x00A0 lor (((code lsr 16)-1) land 0x000F))));
          Buffer.add_char b (char_of_int ((0x0080 lor ((code lsr 10) land 0x003F))));
          Buffer.add_char b (char_of_int 0x00ED);
          Buffer.add_char b (char_of_int ((0x00B0 lor ((code lsr 6) land 0x000F))));
          Buffer.add_char b (char_of_int ((0x0080 lor (code land 0x003F))))
        end)
      u;
    Buffer.contents b
  with _ -> fail_unicode Unable_to_convert_to_modified_utf8

let utf8_of_modified m =
  try
    let b = UTF8.Buf.create 16 in
    let len = String.length m in
    let i = ref 0 in
    while !i < len do
      let code = int_of_char m.[!i] in
      incr i;
      if (code land 0x0080) = 0 then
        UTF8.Buf.add_char b (UChar.chr code)
      else if (code land 0x00E0) = 0x00C0 then begin
        let code' = int_of_char m.[!i] in
        incr i;
        assert ((code' land 0x00C0) = 0x0080);
        let v = ((code land 0x001F) lsl 6) lor (code' land 0x003F) in
        UTF8.Buf.add_char b (UChar.chr v)
      end else if (code land 0x00F0) = 0x00E0 then begin
        let code' = int_of_char m.[!i] in
        incr i;
        let code'' = int_of_char m.[!i] in
        incr i;
        assert ((code' land 0x00C0) = 0x0080);
        assert ((code'' land 0x00C0) = 0x0080);
        let v = ((code land 0x000F) lsl 12)
            lor ((code' land 0x003F) lsl 6)
            lor (code'' land 0x003F) in
        UTF8.Buf.add_char b (UChar.chr v)
      end else begin
        let code' = int_of_char m.[!i] in
        incr i;
        let code'' = int_of_char m.[!i] in
        incr i;
        let code'3 = int_of_char m.[!i] in
        incr i;
        let code'4 = int_of_char m.[!i] in
        incr i;
        let code'5 = int_of_char m.[!i] in
        incr i;
        assert (code = 0x00ED);
        assert ((code' land 0x00F0) = 0x00A0);
        assert ((code'' land 0x00C0) = 0x0080);
        assert (code'3 = 0x00ED);
        assert ((code'4 land 0x00F0) = 0x00B0);
        assert ((code'5 land 0x00C0) = 0x0080);
        let v = 0x010000
            lor ((code' land 0x000F) lsl 16)
            lor ((code'' land 0x003F) lsl 10)
            lor ((code'4 land 0x000F) lsl 6)
            lor (code'5 land 0x003F) in
        UTF8.Buf.add_char b (UChar.chr v)
      end
    done;
    let res = UTF8.Buf.contents b in
    UTF8.validate res;
    res
  with _ -> fail_unicode Unable_to_convert_from_modified_utf8

let string_of_utf8 s =
  try
    let res = Buffer.create (String.length s) in
    UTF8.iter (fun x -> Buffer.add_char res (UChar.char_of x)) s;
    Buffer.contents res
  with _ -> fail_unicode Unable_to_convert_from_utf8

let utf8_of_string s =
  try
    let len = String.length s in
    UTF8.init len (fun i -> UChar.of_char s.[i])
  with _ -> fail_unicode Unable_to_convert_to_utf8

let utf8_length = UTF8.length

let utf8_get s i =
  try
    UTF8.get s i
  with _ -> fail_unicode Invalid_index

let utf8_eq x y =
  (UTF8.compare x y) = 0

let utf8_compare = UTF8.compare

let utf8_index_from s i c =
  let idx = ref (UTF8.move s (UTF8.first s) i) in
  let res = ref i in
  while (not (UTF8.out_of_range s !idx))
      && (not (UChar.eq c (UTF8.look s !idx))) do
    idx := UTF8.next s !idx;
    incr res
  done;
  if UTF8.out_of_range s !idx then
    raise Not_found
  else
    !res

let utf8_rindex_from s i c =
  let idx = ref (UTF8.move s (UTF8.first s) i) in
  let res = ref i in
  while (not (UTF8.out_of_range s !idx))
      && (not (UChar.eq c (UTF8.look s !idx))) do
    idx := UTF8.prev s !idx;
    decr res
  done;
  if UTF8.out_of_range s !idx then
    raise Not_found
  else
    !res

let utf8_substring s first last =
  let len = max 0 (last - first + 1) in
  let res = UTF8.Buf.create len in
  let idx = ref (UTF8.move s (UTF8.first s) first) in
  let i = ref 0 in
  while !i < len do
    if UTF8.out_of_range s !idx then fail_unicode Index_out_of_bounds;
    UTF8.Buf.add_char res (UTF8.look s !idx);
    idx := UTF8.next s !idx;
    incr i
  done;
  UTF8.Buf.contents res

let (++) x y =
  let res = UTF8.Buf.create ((UTF8.length x) + (UTF8.length y)) in
  UTF8.Buf.add_string res x;
  UTF8.Buf.add_string res y;
  UTF8.Buf.contents res

let utf8_concat l =
  let res = UTF8.Buf.create 16 in
  List.iter (UTF8.Buf.add_string res) l;
  UTF8.Buf.contents res

let utf8_concat_sep sep l =
  let res = UTF8.Buf.create 16 in
  (match l with
  | hd :: tl ->
      UTF8.Buf.add_string res hd;
      List.iter (fun x -> UTF8.Buf.add_string res sep; UTF8.Buf.add_string res x) tl
  | [] -> ());
  UTF8.Buf.contents res

let utf8_replace c1 c2 s =
  let res = UTF8.Buf.create (UTF8.length s) in
  UTF8.iter (fun x -> UTF8.Buf.add_char res (if UChar.eq x c1 then c2 else x)) s;
  UTF8.Buf.contents res

let utf8_contains c s =
  try
    ignore (utf8_index_from s 0 c);
    true
  with _ -> false

let utf8_split c s =
  let res = ref [] in
  let curr = UTF8.Buf.create (UTF8.length s) in
  UTF8.iter (fun x ->
    if UChar.eq c x then
      (res := (UTF8.Buf.contents curr) :: !res; UTF8.Buf.clear curr)
    else
      UTF8.Buf.add_char curr x) s;
  let last = UTF8.Buf.contents curr in
  if UTF8.length last > 0 then
    res := last :: !res;
  List.rev !res

external is_printable : char -> bool = "caml_is_printable"

let quote = uchar_of_char '"'

let simple_quote = uchar_of_char '\''

let back_slash = uchar_of_char '\\'

let utf8_escape_delim delim s =
  let res = UTF8.Buf.create ((UTF8.length s) + 2) in
  UTF8.Buf.add_char res delim;
  UTF8.iter
    (fun c ->
      let code = UChar.uint_code c in
      if code > 0x7F || code < 0 then
        let code' = code land 0x0000FFFF in (* from Camomile's UPervasives module *)
        let code'' = code lsr 16 in
        if code'' = 0 then
          UTF8.Buf.add_string res (utf8_of_string (Printf.sprintf "\\u%04X" code'))
        else
          UTF8.Buf.add_string res (utf8_of_string (Printf.sprintf "\\U%04X%04X" code'' code'))
      else
        match char_of_uchar c with (* from stdlib's Char/String modules *)
        | '"' ->
            UTF8.Buf.add_char res back_slash;
            UTF8.Buf.add_char res quote
        | '\\' ->
            UTF8.Buf.add_char res back_slash;
            UTF8.Buf.add_char res back_slash
        | '\n' ->
            UTF8.Buf.add_char res back_slash;
            UTF8.Buf.add_char res (uchar_of_char 'n')
        | '\t' ->
            UTF8.Buf.add_char res back_slash;
            UTF8.Buf.add_char res (uchar_of_char 't')
        | ch ->
            if is_printable ch then
              UTF8.Buf.add_char res (uchar_of_char ch)
            else
              let cc = Char.code ch in
              UTF8.Buf.add_char res back_slash;
              UTF8.Buf.add_char res (uchar_of_char (Char.chr (48 + cc / 100)));
              UTF8.Buf.add_char res (uchar_of_char (Char.chr (48 + (cc / 10) mod 10)));
              UTF8.Buf.add_char res (uchar_of_char (Char.chr (48 + cc mod 10))))
    s;
  UTF8.Buf.add_char res delim;
  UTF8.Buf.contents res

let utf8_escape = utf8_escape_delim quote

let utf8_escape_char c =
  utf8_escape_delim simple_quote (make_utf8 [c])

let utf8_unescape s =
  let len = UTF8.length s in
  if (len < 2) || ((UTF8.get s 0) <> quote) || ((UTF8.get s (pred len)) <> quote) then
    fail_unicode Invalid_escaped_string
  else
    let res = UTF8.Buf.create (len - 2) in
    let idx = ref (UTF8.nth s 1) in
    let last = UTF8.last s in
    let read_digit hex =
      if UTF8.compare_index s !idx last >= 0 then
        fail_unicode Invalid_escaped_string
      else
        let d = Char.uppercase (char_of_uchar (UTF8.look s !idx)) in
        if (d >= '0' && d <= '9') then
          (Char.code d) - (Char.code '0')
        else if (hex && (d >= 'A' && d <= 'F')) then
          (Char.code d) - (Char.code 'A')
        else fail_unicode Invalid_escaped_string in
    while UTF8.compare_index s !idx last < 0 do
      let c = UTF8.look s !idx in
      if uchar_eq back_slash c then begin
        idx := UTF8.next s !idx;
        if UTF8.compare_index s !idx last >= 0 then
          fail_unicode Invalid_escaped_string
        else
          let c' = UTF8.look s !idx in
          if uchar_eq quote c' then
            (UTF8.Buf.add_char res quote; idx := UTF8.next s !idx)
          else if uchar_eq back_slash c' then
            (UTF8.Buf.add_char res back_slash; idx := UTF8.next s !idx)
          else if uchar_eq (uchar_of_char 'n') c' then
            (UTF8.Buf.add_char res (uchar_of_char '\n'); idx := UTF8.next s !idx)
          else if uchar_eq (uchar_of_char 't') c' then
            (UTF8.Buf.add_char res (uchar_of_char '\t'); idx := UTF8.next s !idx)
          else if uchar_eq (uchar_of_char 'u') c' then
            let a = read_digit true in
            let b = read_digit true in
            let c = read_digit true in
            let d = read_digit true in
            let code = (a lsl 12) + (b lsl 8) + (c lsl 4) + d in
            UTF8.Buf.add_char res (UChar.chr_of_uint code)
          else if uchar_eq (uchar_of_char 'U') c' then
            let a = read_digit true in
            let b = read_digit true in
            let c = read_digit true in
            let d = read_digit true in
            let e = read_digit true in
            let f = read_digit true in
            let g = read_digit true in
            let h = read_digit true in
            let code = (a lsl 28) + (b lsl 24) + (c lsl 20) + (d lsl 16)
              + (e lsl 12) + (f lsl 8) + (g lsl 4) + h in
            UTF8.Buf.add_char res (UChar.chr_of_uint code)
          else (* digits *)
            let x = Char.code (char_of_uchar c') in
            idx := UTF8.next s !idx;
            let y = read_digit false in
            let z = read_digit false in
            let code = ((x - 48) * 100) + ((y - 48) * 10) + (z - 48) in
            UTF8.Buf.add_char res (uchar_of_char (Char.chr code))
      end else begin
        UTF8.Buf.add_char res c;
        idx := UTF8.next s !idx
      end
    done;
    UTF8.Buf.contents res


(* UTF8 hash tables *)

module UTF8HashedType = struct
  type t = utf8
  let equal = utf8_eq
  let hash x = Hashtbl.hash (string_of_utf8 x)
end

module UTF8Hashtbl = Hashtbl.Make (UTF8HashedType)


(* UTF8 buffers *)

type utf8_buf = UTF8.Buf.buf

let utf8_buf_make () = UTF8.Buf.create 16

let utf8_buf_add_char = UTF8.Buf.add_char

let utf8_buf_add_string = UTF8.Buf.add_string

let eol = uchar_of_char '\n'

let utf8_buf_add_endline b s =
  utf8_buf_add_string b s;
  utf8_buf_add_char b eol

let utf8_buf_contents =  UTF8.Buf.contents


(* UTF8 line-oriented readers *)

module Encoding = CharEncoding.Configure (CamomileDefaultConfig)

module UTF8Line = ULine.Make (UTF8)

type utf8_line_reader = UTF8Line.input_line

let utf8_line_reader_make ic =
  let ch = new Encoding.in_channel Encoding.utf8 ic in
  new UTF8Line.input_line ch

let utf8_line_reader_get lr = lr#get ()

let utf8_line_reader_close lr = lr#close_in ()


(* UTF8 line-oriented writers *)

type utf8_line_writer = UTF8Line.output_line

let utf8_line_writer_make oc =
  let ch = new Encoding.out_channel Encoding.utf8 oc in
  new UTF8Line.output_line ch

let utf8_line_writer_put lw s = lw#put s

let utf8_line_writer_flush lw = lw#flush ()

let utf8_line_writer_close lw = lw#close_out ()


(* State objects for lexers over UTF8 strings *)

class lexer_state s =
  object (self)
    val mutable next = UTF8.first s
    method is_available = not (UTF8.out_of_range s next)
    method look_ahead_list l =
      if UTF8.out_of_range s next then
        fail_lexer End_of_lexer
      else
        let next_char = UTF8.look s next in
        List.exists (UChar.eq next_char) l
    method look_ahead ch =
      self#look_ahead_list [ch]
    method consume_char =
      if UTF8.out_of_range s next then
        fail_lexer End_of_lexer
      else
        let res = UTF8.look s next in next <- UTF8.next s next; res
    method consume =
      ignore (self#consume_char)
    method consume_only ch =
      let next_char = self#consume_char in
      if not (UChar.eq ch next_char) then
        fail_lexer (Invalid_consume((char_of_uchar ch), (char_of_uchar next_char)))
    method consume_until_list l =
      let b = UTF8.Buf.create 16 in
      while not (self#look_ahead_list l) do
        UTF8.Buf.add_char b (UTF8.look s next);
        next <- UTF8.next s next
      done;
      utf8_buf_contents b
    method consume_until ch =
      self#consume_until_list [ch]
  end
