(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)


(* Type *)

type t =
  | Java_1_0
  | Java_1_1
  | Java_1_2
  | Java_1_3
  | Java_1_4
  | Java_1_5
  | Java_1_6
  | Java_1_7

type bound = {
    bound_version : t;
    bound_feature : string;
  }

type bounds = bound * bound option

let make_bounds f lo hi =
  { bound_version = lo; bound_feature = String.copy f; },
  (match hi with
  | Some v -> Some { bound_version = v; bound_feature = String.copy f; }
  | None -> None)


(* Exception *)

type error =
  | Invalid_version
  | Unsupported_feature of (int * int) * string
  | Deprecated_feature of (int * int) * string

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_version -> "invalid version"
  | Unsupported_feature ((mj, mn), f) -> Printf.sprintf "%s is not supported by class file version %d.%d" f mj mn
  | Deprecated_feature ((mj, mn), f) -> Printf.sprintf "%s is deprecated after class file version %d.%d" f mj mn

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Constants *)

let min_supported = 45, 0

let max_supported = 51, 0


(* Conversion functions *)

let major_minor_of_version = function
  | Java_1_0 -> 45, 3
  | Java_1_1 -> 45, 4
  | Java_1_2 -> 46, 0
  | Java_1_3 -> 47, 0
  | Java_1_4 -> 48, 0
  | Java_1_5 -> 49, 0
  | Java_1_6 -> 50, 0
  | Java_1_7 -> 51, 0

let version_of_major_minor = function
  | 45, x when x >= 0 && x <= 3 -> Java_1_0
  | 45, x when x >= 0 && x <= 65535 -> Java_1_1
  | 46, 0 -> Java_1_2
  | 46, _ | 47, 0 -> Java_1_3
  | 47, _ | 48, 0 -> Java_1_4
  | 48, _ | 49, 0 -> Java_1_5
  | 49, _ | 50, 0 -> Java_1_6
  | 50, _ | 51, 0 -> Java_1_7
  | _ -> fail Invalid_version


(* Utility functions *)

let at_least f v x =
  if x < v then
    let v' = major_minor_of_version v in
    let x' = major_minor_of_version x in
    Printf.printf "x = %d, %d and v = %d, %d\n" (fst x') (snd x') (fst v') (snd v');
    fail (Unsupported_feature (v', f))

let at_most f v x =
  if x > v then
    let v' = major_minor_of_version v in
    fail (Deprecated_feature (v', f))

let check (inf, sup) x =
  at_least inf.bound_feature inf.bound_version x;
  match sup with
  | Some s -> at_most s.bound_feature s.bound_version x
  | None -> ()

let intersect (min_v1, max_v1) (min_v2, max_v2) =
  let copy b =
    { bound_version = b.bound_version;
      bound_feature = String.copy b.bound_feature; } in
  let min_res = (* max of min_v1 and min_v2 *)
    if min_v1.bound_version >= min_v2.bound_version then
      copy min_v1
    else
      copy min_v2 in
  let max_res =
    match max_v1, max_v2 with
    | None, None -> None
    | Some x, None -> Some (copy x)
    | None, Some x -> Some (copy x)
    | Some x, Some y -> (* min of max_v1 and max_v2 *)
        if x.bound_version <= y.bound_version then
          Some (copy x)
        else
          Some (copy y) in
  min_res, max_res

let intersect_list = function
  | hd :: tl -> List.fold_left intersect hd tl
  | [] -> invalid_arg "Version.intersect_list"
