/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    private static final MethodRef println =
	new MethodRef("java.io.PrintStream",
		      "println",
		      Descriptor.VOID,
		      descs(new Descriptor.Class("java.lang.String")));

    public static void main(final String[] args) throws Throwable {
	final List<Instruction> sumFloatsInstructions =
	    asList(FLOAD_0,
		   FLOAD_1,
		   FLOAD_2,
		   FLOAD_3,
		   new FLOAD((short) 4),
		   FADD,
		   FADD,
		   FADD,
		   FADD,
		   FRETURN);
	final List<Instruction> sumDoublesInstructions =
	    asList(DLOAD_0,
		   DLOAD_2,
		   new DLOAD((short) 4),
		   DADD,
		   DADD,
		   DRETURN);
	final List<Instruction> sumIntsInstructions =
	    asList(ILOAD_0,
		   ILOAD_1,
		   ILOAD_2,
		   ILOAD_3,
		   new ILOAD((short) 4),
		   new ILOAD((short) 5),
		   new ILOAD((short) 6),
		   IADD,
		   IADD,
		   IADD,
		   IADD,
		   IADD,
		   IADD,
		   IRETURN);
	final List<Instruction> sumLongsInstructions =
	    asList(LLOAD_0,
		   LLOAD_2,
		   new LLOAD((short) 4),
		   LADD,
		   LADD,
		   LRETURN);
	final List<Instruction> instructions =
	    asList(/* floats */
		   new GETSTATIC(systOut),
		   FCONST_0,
		   FCONST_1,
		   FCONST_1,
		   FCONST_2,
		   FCONST_2,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "sum",
						  Descriptor.FLOAT,
						  descs(Descriptor.FLOAT,
							Descriptor.FLOAT,
							Descriptor.FLOAT,
							Descriptor.FLOAT,
							Descriptor.FLOAT))),
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* doubles */
		   new GETSTATIC(systOut),
		   DCONST_0,
		   DCONST_1,
		   DCONST_1,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "sum",
						  Descriptor.DOUBLE,
						  descs(Descriptor.DOUBLE,
							Descriptor.DOUBLE,
							Descriptor.DOUBLE))),
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* ints */
		   new GETSTATIC(systOut),
		   ICONST_0,
		   ICONST_1,
		   ICONST_2,
		   ICONST_3,
		   ICONST_4,
		   ICONST_5,
		   ICONST_M1,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "sum",
						  Descriptor.INT,
						  descs(Descriptor.INT,
							Descriptor.INT,
							Descriptor.INT,
							Descriptor.INT,
							Descriptor.INT,
							Descriptor.INT,
							Descriptor.INT))),
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* longs */
		   new GETSTATIC(systOut),
		   LCONST_0,
		   LCONST_1,
		   LCONST_1,
		   new INVOKESTATIC(new MethodRef("pack.Test",
						  "sum",
						  Descriptor.LONG,
						  descs(Descriptor.LONG,
							Descriptor.LONG,
							Descriptor.LONG))),
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),
		   RETURN);
	final Method sumFloats =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "sum",
			  Descriptor.FLOAT,
			  descs(Descriptor.FLOAT,
				Descriptor.FLOAT,
				Descriptor.FLOAT,
				Descriptor.FLOAT,
				Descriptor.FLOAT),
			  5,
			  5,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  sumFloatsInstructions);
	final Method sumDoubles =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "sum",
			  Descriptor.DOUBLE,
			  descs(Descriptor.DOUBLE,
				Descriptor.DOUBLE,
				Descriptor.DOUBLE),
			  6,
			  6,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  sumDoublesInstructions);
	final Method sumInts =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "sum",
			  Descriptor.INT,
			  descs(Descriptor.INT,
				Descriptor.INT,
				Descriptor.INT,
				Descriptor.INT,
				Descriptor.INT,
				Descriptor.INT,
				Descriptor.INT),
			  7,
			  7,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  sumIntsInstructions);
	final Method sumLongs =
	    compileMethod(asList(AccessFlag.Private, AccessFlag.Static),
			  "sum",
			  Descriptor.LONG,
			  descs(Descriptor.LONG,
				Descriptor.LONG,
				Descriptor.LONG),
			  6,
			  6,
			  Collections.<Attribute.ExceptionTableElement>emptyList(),
			  Collections.<Attribute>emptyList(),
			  Collections.<Attribute>emptyList(),
			  sumLongsInstructions);
	final Method main = compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
					  "main",
					  Descriptor.VOID,
					  descs(new Descriptor.Array(new Descriptor.Class("java.lang.String"))),
					  8,
					  4,
					  Collections.<Attribute.ExceptionTableElement>emptyList(),
					  Collections.<Attribute>emptyList(),
					  Collections.<Attribute>emptyList(),
					  instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						null,
						null,
						asList(sumFloats,
						       sumDoubles,
						       sumInts,
						       sumLongs,
						       main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
